"""Nicer config class.
"""

import os
import os.path
import re
import socket
from configparser import MAX_INTERPOLATION_DEPTH, ConfigParser
from configparser import Error as ConfigError
from configparser import (
    ExtendedInterpolation, Interpolation, InterpolationDepthError,
    InterpolationError, NoOptionError, NoSectionError, RawConfigParser,
)
from typing import Dict, List, Mapping, Optional, Sequence, Tuple, MutableMapping, Set

import skytools

__all__ = (
    'Config', 'NoOptionError', 'ConfigError',
    'ConfigParser', 'ExtendedConfigParser', 'ExtendedCompatConfigParser',
    'InterpolationError', 'NoOptionError', 'NoSectionError',
)


def read_versioned_config(filenames: Sequence[str], main_section: str) -> ConfigParser:
    """Pick syntax based on "config_format" value.
    """
    rcf = RawConfigParser()
    rcf.read(filenames)

    # avoid has_option here, so value can live in DEFAULT section
    ver = rcf.get(main_section, "config_format", fallback="1")
    if ver == "1":
        cf = ConfigParser()
    elif ver == "2":
        cf = ExtendedConfigParser()
    else:
        raise ConfigError('Unsupported config format %r in %r' % (ver, filenames))
    cf.read(filenames)
    return cf


class Config:
    """Bit improved ConfigParser.

    Additional features:
     - Remembers section.
     - Accepts defaults in get() functions.
     - List value support.
    """
    main_section: str               # main section
    filename: Optional[str]         # file name that was loaded
    override: Mapping[str, str]     # override values in config file
    defs: Mapping[str, str]         # defaults visible in all sections
    cf: ConfigParser                # actual ConfigParser instance

    def __init__(self, main_section: str,
                 filename: Optional[str],
                 sane_config: Optional[bytes] = None,   # unused
                 user_defs: Optional[Mapping[str, str]] = None,
                 override: Optional[Mapping[str, str]] = None,
                 ignore_defs: bool = False) -> None:
        """Initialize Config and read from file.
        """
        # use config file name as default job_name
        if filename:
            job_name = os.path.splitext(os.path.basename(filename))[0]
        else:
            job_name = main_section

        # initialize defaults, make them usable in config file
        if ignore_defs:
            self.defs = {}
        else:
            self.defs = {
                'job_name': job_name,
                'service_name': main_section,
                'host_name': socket.gethostname(),
            }
            if filename:
                self.defs['config_dir'] = os.path.dirname(filename)
                self.defs['config_file'] = filename
            if user_defs:
                self.defs.update(user_defs)

        self.main_section = main_section
        self.filename = filename
        self.override = override or {}

        if filename is None:
            self.cf = ConfigParser()
            self.cf.add_section(main_section)
        elif not os.path.isfile(filename):
            raise ConfigError('Config file not found: ' + filename)
        else:
            self.cf = read_versioned_config([filename], main_section)

        self.reload()

    def reload(self) -> None:
        """Re-reads config file."""
        if self.filename:
            self.cf.read(self.filename)
        if not self.cf.has_section(self.main_section):
            raise NoSectionError(self.main_section)

        # apply default if key not set
        for k, v in self.defs.items():
            if not self.cf.has_option(self.main_section, k):
                self.cf.set(self.main_section, k, v)

        # apply overrides
        if self.override:
            for k, v in self.override.items():
                self.cf.set(self.main_section, k, v)

    def get(self, key: str, default: Optional[str] = None) -> str:
        """Reads string value, if not set then default."""

        if not self.cf.has_option(self.main_section, key):
            if default is None:
                raise NoOptionError(key, self.main_section)
            return default

        return str(self.cf.get(self.main_section, key))

    def getint(self, key: str, default: Optional[int] = None) -> int:
        """Reads int value, if not set then default."""

        if not self.cf.has_option(self.main_section, key):
            if default is None:
                raise NoOptionError(key, self.main_section)
            return default

        return self.cf.getint(self.main_section, key)

    def getboolean(self, key: str, default: Optional[bool] = None) -> bool:
        """Reads boolean value, if not set then default."""

        if not self.cf.has_option(self.main_section, key):
            if default is None:
                raise NoOptionError(key, self.main_section)
            return default

        return self.cf.getboolean(self.main_section, key)

    def getfloat(self, key: str, default: Optional[float] = None) -> float:
        """Reads float value, if not set then default."""

        if not self.cf.has_option(self.main_section, key):
            if default is None:
                raise NoOptionError(key, self.main_section)
            return default

        return self.cf.getfloat(self.main_section, key)

    def getlist(self, key: str, default: Optional[List[str]] = None) -> List[str]:
        """Reads comma-separated list from key."""

        if not self.cf.has_option(self.main_section, key):
            if default is None:
                raise NoOptionError(key, self.main_section)
            return default

        s = self.get(key).strip()
        res: List[str] = []
        if not s:
            return res
        for v in s.split(","):
            res.append(v.strip())
        return res

    def getdict(self, key: str, default: Optional[Dict[str, str]] = None) -> Dict[str, str]:
        """Reads key-value dict from parameter.

        Key and value are separated with ':'.  If missing,
        key itself is taken as value.
        """

        if not self.cf.has_option(self.main_section, key):
            if default is None:
                raise NoOptionError(key, self.main_section)
            return default

        s = self.get(key).strip()
        res: Dict[str, str] = {}
        if not s:
            return res
        for kv in s.split(","):
            tmp = kv.split(':', 1)
            if len(tmp) > 1:
                k = tmp[0].strip()
                v = tmp[1].strip()
            else:
                k = kv.strip()
                v = k
            res[k] = v
        return res

    def getfile(self, key: str, default: Optional[str] = None) -> str:
        """Reads filename from config.

        In addition to reading string value, expands ~ to user directory.
        """
        fn = self.get(key, default)
        if fn == "" or fn == "-":
            return fn
        # simulate that the cwd is script location
        #path = os.path.dirname(sys.argv[0])
        #  seems bad idea, cwd should be cwd

        fn = os.path.expanduser(fn)

        return fn

    def getbytes(self, key: str, default: Optional[str] = None) -> int:
        """Reads a size value in human format, if not set then default.

        Examples: 1, 2 B, 3K, 4 MB
        """

        if not self.cf.has_option(self.main_section, key):
            if default is None:
                raise NoOptionError(key, self.main_section)
            s = default
        else:
            s = self.cf.get(self.main_section, key)

        return skytools.hsize_to_bytes(s)

    def get_wildcard(self, key: str, values: Sequence[str] = (), default: Optional[str] = None) -> str:
        """Reads a wildcard property from conf and returns its string value, if not set then default."""

        orig_key = key
        keys = [key]

        for wild in values:
            key = key.replace('*', wild, 1)
            keys.append(key)
        keys.reverse()

        for k in keys:
            if self.cf.has_option(self.main_section, k):
                return self.cf.get(self.main_section, k)

        if default is None:
            raise NoOptionError(orig_key, self.main_section)
        return default

    def sections(self) -> Sequence[str]:
        """Returns list of sections in config file, excluding DEFAULT."""
        return self.cf.sections()

    def has_section(self, section: str) -> bool:
        """Checks if section is present in config file, excluding DEFAULT."""
        return self.cf.has_section(section)

    def clone(self, main_section: str) -> "Config":
        """Return new Config() instance with new main section on same config file."""
        return Config(main_section, self.filename)

    def options(self) -> Sequence[str]:
        """Return list of options in main section."""
        return self.cf.options(self.main_section)

    def has_option(self, opt: str) -> bool:
        """Checks if option exists in main section."""
        return self.cf.has_option(self.main_section, opt)

    def items(self) -> Sequence[Tuple[str, str]]:
        """Returns list of (name, value) for each option in main section."""
        return self.cf.items(self.main_section)

    # define some aliases (short-cuts / backward compatibility cruft)
    getbool = getboolean


ParserSection = Mapping[str, str]
ParserState = MutableMapping[str, ParserSection]
#ParserState = ConfigParser
ParserLoop = Set[Tuple[str, str]]

class ExtendedInterpolationCompat(Interpolation):
    _EXT_VAR_RX = r'\$\$|\$\{[^(){}]+\}'
    _OLD_VAR_RX = r'%%|%\([^(){}]+\)s'
    _var_rc = re.compile('(%s|%s)' % (_EXT_VAR_RX, _OLD_VAR_RX))
    _bad_rc = re.compile('[%$]')

    def before_get(self, parser: ParserState, section: str, option: str, value: str, defaults: ParserSection) -> str:
        dst: List[str] = []
        self._interpolate_ext(dst, parser, section, option, value, defaults, set())
        return ''.join(dst)

    def before_set(self, parser: ParserState, section: str, option: str, value: str) -> str:
        sub = self._var_rc.sub('', value)
        if self._bad_rc.search(sub):
            raise ValueError("invalid interpolation syntax in %r" % value)
        return value

    def _interpolate_ext(self, dst: List[str], parser: ParserState, section: str, option: str, rawval: str, defaults: ParserSection, loop_detect: ParserLoop) -> None:
        if not rawval:
            return

        if len(loop_detect) > MAX_INTERPOLATION_DEPTH:
            raise InterpolationDepthError(option, section, rawval)

        xloop = (section, option)
        if xloop in loop_detect:
            raise InterpolationError(section, option, 'Loop detected: %r in %r' % (xloop, loop_detect))
        loop_detect.add(xloop)

        parts = self._var_rc.split(rawval)
        for i, frag in enumerate(parts):
            fullkey = None
            use_vars: Optional[ParserSection] = defaults
            if i % 2 == 0:
                dst.append(frag)
                continue
            if frag in ('$$', '%%'):
                dst.append(frag[0])
                continue
            if frag.startswith('${') and frag.endswith('}'):
                fullkey = frag[2:-1]

                # use section access only for new-style keys
                if ':' in fullkey:
                    ksect, key = fullkey.split(':', 1)
                    use_vars = None
                else:
                    ksect, key = section, fullkey
            elif frag.startswith('%(') and frag.endswith(')s'):
                fullkey = frag[2:-2]
                ksect, key = section, fullkey
            else:
                raise InterpolationError(section, option, 'Internal parse error: %r' % frag)

            if isinstance(parser, RawConfigParser):
                key = parser.optionxform(key)
                newpart = parser.get(ksect, key, raw=True, vars=use_vars)
                if newpart is None:
                    raise InterpolationError(ksect, key, 'Key referenced is None')
                self._interpolate_ext(dst, parser, ksect, key, newpart, defaults, loop_detect)

        loop_detect.remove(xloop)


class ExtendedConfigParser(ConfigParser):
    """ConfigParser that uses Python3-style extended interpolation by default.

    Syntax: ${var} and ${section:var}
    """
    _DEFAULT_INTERPOLATION: Interpolation = ExtendedInterpolation()


class ExtendedCompatConfigParser(ExtendedConfigParser):
    r"""Support both extended "${}" syntax from python3 and old "%()s" too.

    New ${} syntax allows ${key} to refer key in same section,
    and ${sect:key} to refer key in other sections.
    """
    _DEFAULT_INTERPOLATION: Interpolation = ExtendedInterpolationCompat()





from skytools.dbservice import transform_fields


def test_transform_fields() -> None:
    rows = []
    rows.append({'time': '22:00', 'metric': 'count', 'value': 100})
    rows.append({'time': '22:00', 'metric': 'dur', 'value': 7})
    rows.append({'time': '23:00', 'metric': 'count', 'value': 200})
    rows.append({'time': '23:00', 'metric': 'dur', 'value': 5})
    res = list(transform_fields(rows, ['time'], 'metric', 'value'))
    assert res[0] == {'count': 100, 'dur': 7, 'time': '22:00'}
    assert res[1] == {'count': 200, 'dur': 5, 'time': '23:00'}




"""PLPY helper module for applying row events from pgq.logutriga().
"""

from typing import Sequence, Optional

import skytools

try:
    import plpy
except ImportError:
    pass


## TODO: automatic fkey detection
# find FK columns
FK_SQL = """
SELECT (SELECT array_agg( (SELECT attname::text FROM pg_attribute
                            WHERE attrelid = conrelid AND attnum = conkey[i]))
          FROM generate_series(1, array_upper(conkey, 1)) i) AS kcols,
       (SELECT array_agg( (SELECT attname::text FROM pg_attribute
                            WHERE attrelid = confrelid AND attnum = confkey[i]))
          FROM generate_series(1, array_upper(confkey, 1)) i) AS fcols,
       confrelid::regclass::text AS ftable
  FROM pg_constraint
 WHERE conrelid = {tbl}::regclass AND contype='f'
"""


class DataError(Exception):
    "Invalid data"


def colfilter_full(rnew, rold):
    return rnew


def colfilter_changed(rnew, rold):
    res = {}
    for k, _ in rnew:
        if rnew[k] != rold[k]:
            res[k] = rnew[k]
    return res


def canapply_dummy(rnew, rold):
    return True


def canapply_tstamp_helper(rnew, rold, tscol):
    tnew = rnew[tscol]
    told = rold[tscol]
    if not tnew[0].isdigit():
        raise DataError('invalid timestamp')
    if not told[0].isdigit():
        raise DataError('invalid timestamp')
    return tnew > told


def applyrow(tblname, ev_type, new_row,
             backup_row=None,
             alt_pkey_cols=None,
             fkey_cols=None,
             fkey_ref_table=None,
             fkey_ref_cols=None,
             fn_canapply=canapply_dummy,
             fn_colfilter=colfilter_full):
    """Core logic.  Actual decisions will be done in callback functions.

    - [IUD]: If row referenced by fkey does not exist, event is not applied
    - If pkey does not exist but alt_pkey does, row is not applied.

    @param tblname: table name, schema-qualified
    @param ev_type: [IUD]:pkey1,pkey2
    @param alt_pkey_cols: list of alternatice columns to consuder
    @param fkey_cols: columns in this table that refer to other table
    @param fkey_ref_table: other table referenced here
    @param fkey_ref_cols: column in other table that must match
    @param fn_canapply: callback function, gets new and old row, returns whether the row should be applied
    @param fn_colfilter: callback function, gets new and old row, returns dict of final columns to be applied
    """

    gd = None

    # parse ev_type
    tmp = ev_type.split(':', 1)
    if len(tmp) != 2 or tmp[0] not in ('I', 'U', 'D'):
        raise DataError('Unsupported ev_type: ' + repr(ev_type))
    if not tmp[1]:
        raise DataError('No pkey in event')

    cmd = tmp[0]
    pkey_cols = tmp[1].split(',')
    qtblname = skytools.quote_fqident(tblname)

    # parse ev_data
    fields = skytools.db_urldecode(new_row)

    if ev_type.find('}') >= 0:
        raise DataError('Really suspicious activity')
    if ",".join(fields.keys()).find('}') >= 0:
        raise DataError('Really suspicious activity 2')

    # generate pkey expressions
    tmp = ["%s = {%s}" % (skytools.quote_ident(k), k) for k in pkey_cols]
    pkey_expr = " and ".join(tmp)
    alt_pkey_expr = None
    if alt_pkey_cols:
        tmp = ["%s = {%s}" % (skytools.quote_ident(k), k) for k in alt_pkey_cols]
        alt_pkey_expr = " and ".join(tmp)

    log = "data ok"

    #
    # Row data seems fine, now apply it
    #

    res: Optional[Sequence[skytools.dbdict]]
    oldrow: Optional[skytools.dbdict]

    if fkey_ref_table:
        tmp = []
        for k, rk in zip(fkey_cols, fkey_ref_cols):
            tmp.append("%s = {%s}" % (skytools.quote_ident(rk), k))
        fkey_expr = " and ".join(tmp)
        q = "select 1 from only %s where %s" % (
            skytools.quote_fqident(fkey_ref_table),
            fkey_expr)
        res = skytools.plpy_exec(gd, q, fields)
        if not res:
            return "IGN: parent row does not exist"
        log += ", fkey ok"

    # fetch old row
    if alt_pkey_expr:
        q = "select * from only %s where %s for update" % (qtblname, alt_pkey_expr)
        res = skytools.plpy_exec(gd, q, fields)
        if res:
            oldrow = res[0]
            # if altpk matches, but pk not, then delete
            need_del = 0
            for k in pkey_cols:
                # fixme: proper type cmp?
                if fields[k] != str(oldrow[k]):
                    need_del = 1
                    break
            if need_del:
                log += ", altpk del"
                q = "delete from only %s where %s" % (qtblname, alt_pkey_expr)
                skytools.plpy_exec(gd, q, fields)
                res = None
            else:
                log += ", altpk ok"
    else:
        # no altpk
        q = "select * from only %s where %s for update" % (qtblname, pkey_expr)
        res = skytools.plpy_exec(None, q, fields)

    # got old row, with same pk and altpk
    if res:
        oldrow = res[0]
        log += ", old row"
        ok = fn_canapply(fields, oldrow)
        if ok:
            log += ", new row better"
        if not ok:
            # ignore the update
            return "IGN:" + log + ", current row more up-to-date"
    else:
        log += ", no old row"
        oldrow = None

    if res:
        if cmd == 'I':
            cmd = 'U'
    else:
        if cmd == 'U':
            cmd = 'I'

    # allow column changes
    if oldrow:
        fields2 = fn_colfilter(fields, oldrow)
        for k in pkey_cols:
            if k not in fields2:
                fields2[k] = fields[k]
        fields = fields2

    # apply change
    if cmd == 'I':
        q = skytools.mk_insert_sql(fields, tblname, pkey_cols)
    elif cmd == 'U':
        q = skytools.mk_update_sql(fields, tblname, pkey_cols)
    elif cmd == 'D':
        q = skytools.mk_delete_sql(fields, tblname, pkey_cols)
    else:
        plpy.error('Huh')

    plpy.execute(q)

    return log


def ts_conflict_handler(gd, args):
    """Conflict handling based on timestamp column."""

    conf = skytools.db_urldecode(args[0])
    timefield = conf['timefield']
    ev_type = args[1]
    ev_data = args[2]
    ev_extra1 = args[3]
    ev_extra2 = args[4]
    #ev_extra3 = args[5]
    #ev_extra4 = args[6]
    altpk = None
    cf_altpk = conf.get('altpk')
    if cf_altpk:
        altpk = cf_altpk.split(',')

    def ts_canapply(rnew, rold):
        return canapply_tstamp_helper(rnew, rold, timefield)

    return applyrow(
        ev_extra1, ev_type, ev_data,
        backup_row=ev_extra2,
        alt_pkey_cols=altpk,
        fkey_ref_table=conf.get('fkey_ref_table'),
        fkey_ref_cols=conf.get('fkey_ref_cols'),
        fkey_cols=conf.get('fkey_cols'),
        fn_canapply=ts_canapply
    )




"""Our log handlers for Python's logging package.
"""

import logging
import logging.handlers
import os
import socket
import time
from logging import LoggerAdapter

from typing import Any, Optional, Union, Dict

import skytools
import skytools.tnetstrings
from skytools.basetypes import Buffer

__all__ = ['getLogger']

# add TRACE level
TRACE = 5
logging.TRACE = TRACE   # type: ignore
logging.addLevelName(TRACE, 'TRACE')

# extra info to be added to each log record
_service_name = 'unknown_svc'
_job_name = 'unknown_job'
_hostname = socket.gethostname()
try:
    _hostaddr = socket.gethostbyname(_hostname)
except BaseException:
    _hostaddr = "0.0.0.0"
_log_extra = {
    'job_name': _job_name,
    'service_name': _service_name,
    'hostname': _hostname,
    'hostaddr': _hostaddr,
}


def set_service_name(service_name: str, job_name: str) -> None:
    """Set info about current script."""
    global _service_name, _job_name

    _service_name = service_name
    _job_name = job_name

    _log_extra['job_name'] = _job_name
    _log_extra['service_name'] = _service_name


# Make extra fields available to all log records
_old_factory = logging.getLogRecordFactory()


def _new_factory(*args: Any, **kwargs: Any) -> logging.LogRecord:
    record = _old_factory(*args, **kwargs)
    record.__dict__.update(_log_extra)
    return record


logging.setLogRecordFactory(_new_factory)


# configurable file logger
class EasyRotatingFileHandler(logging.handlers.RotatingFileHandler):
    """Easier setup for RotatingFileHandler."""
    def __init__(self, filename: str, maxBytes: int = 10 * 1024 * 1024, backupCount: int = 3) -> None:
        """Args same as for RotatingFileHandler, but in filename '~' is expanded."""
        fn = os.path.expanduser(filename)
        super().__init__(fn, maxBytes=maxBytes, backupCount=backupCount)


# send JSON message over UDP
class UdpLogServerHandler(logging.handlers.DatagramHandler):
    """Sends log records over UDP to logserver in JSON format."""

    # map logging levels to logserver levels
    _level_map = {
        logging.DEBUG: 'DEBUG',
        logging.INFO: 'INFO',
        logging.WARNING: 'WARN',
        logging.ERROR: 'ERROR',
        logging.CRITICAL: 'FATAL',
    }

    # JSON message template
    _log_template = '{\n\t'\
        '"logger": "skytools.UdpLogServer",\n\t'\
        '"timestamp": %.0f,\n\t'\
        '"level": "%s",\n\t'\
        '"thread": null,\n\t'\
        '"message": %s,\n\t'\
        '"properties": {"application":"%s", "apptype": "%s", "type": "sys", "hostname":"%s", "hostaddr": "%s"}\n'\
        '}\n'

    # cut longer msgs
    MAXMSG = 1024

    def makePickle(self, record: logging.LogRecord) -> bytes:
        """Create message in JSON format."""
        # get & cut msg
        msg = self.format(record)
        if len(msg) > self.MAXMSG:
            msg = msg[:self.MAXMSG]
        txt_level = self._level_map.get(record.levelno, "ERROR")
        hostname = _hostname
        hostaddr = _hostaddr
        jobname = _job_name
        svcname = _service_name
        pkt = self._log_template % (
            time.time() * 1000, txt_level, skytools.quote_json(msg),
            jobname, svcname, hostname, hostaddr
        )
        return pkt.encode("utf8")

    def send(self, s: Buffer) -> None:
        """Disable socket caching."""
        sock = self.makeSocket()
        sock.sendto(s, (self.host, self.port))
        sock.close()


# send TNetStrings message over UDP
class UdpTNetStringsHandler(logging.handlers.DatagramHandler):
    """ Sends log records in TNetStrings format over UDP. """

    # LogRecord fields to send
    send_fields = [
        'created', 'exc_text', 'levelname', 'levelno', 'message', 'msecs', 'name',
        'hostaddr', 'hostname', 'job_name', 'service_name']

    _udp_reset: float = 0

    def makePickle(self, record: logging.LogRecord) -> bytes:
        """ Create message in TNetStrings format.
        """
        msg = {}
        self.format(record)  # render 'message' attribute and others
        for k in self.send_fields:
            msg[k] = record.__dict__[k]
        tnetstr = skytools.tnetstrings.dumps(msg)
        return tnetstr

    def send(self, s: Buffer) -> None:
        """ Cache socket for a moment, then recreate it.
        """
        now = time.time()
        if now - 1 > self._udp_reset:
            if self.sock:
                self.sock.close()
            self.sock = self.makeSocket()
            self._udp_reset = now
        if self.sock:
            self.sock.sendto(s, (self.host, self.port))


class LogDBHandler(logging.handlers.SocketHandler):
    """Sends log records into PostgreSQL server.

    Additionally, does some statistics aggregating,
    to avoid overloading log server.

    It subclasses SocketHandler to get throtthling for
    failed connections.
    """

    # map codes to string
    _level_map = {
        logging.DEBUG: 'DEBUG',
        logging.INFO: 'INFO',
        logging.WARNING: 'WARNING',
        logging.ERROR: 'ERROR',
        logging.CRITICAL: 'FATAL',
    }

    sock: Any
    closeOnError: bool
    connect_string: str
    stat_cache: Dict[str, Union[int, float]]
    stat_flush_period: float
    last_stat_flush: float

    def __init__(self, connect_string: str) -> None:
        """
        Initializes the handler with a specific connection string.
        """

        super().__init__("localhost", 1)
        self.closeOnError = True

        self.connect_string = connect_string

        self.stat_cache = {}
        self.stat_flush_period = 60
        # send first stat line immediately
        self.last_stat_flush = 0

    def createSocket(self) -> None:
        try:
            super().createSocket()
        except BaseException:
            self.sock = self.makeSocket()

    def makeSocket(self, timeout: float = 1) -> Any:
        """Create server connection.
        In this case its not socket but database connection."""

        db = skytools.connect_database(self.connect_string)
        db.set_isolation_level(0)  # autocommit
        return db

    def emit(self, record: logging.LogRecord) -> None:
        """Process log record."""

        # we do not want log debug messages
        if record.levelno < logging.INFO:
            return

        try:
            self.process_rec(record)
        except (SystemExit, KeyboardInterrupt):
            raise
        except BaseException:
            self.handleError(record)

    def process_rec(self, record: logging.LogRecord) -> None:
        """Aggregate stats if needed, and send to logdb."""
        # render msg
        msg = self.format(record)

        # dont want to send stats too ofter
        if record.levelno == logging.INFO and msg and msg[0] == "{":
            self.aggregate_stats(msg)
            if time.time() - self.last_stat_flush >= self.stat_flush_period:
                self.flush_stats(_job_name)
            return

        if record.levelno < logging.INFO:
            self.flush_stats(_job_name)

        # dont send more than one line
        ln = msg.find('\n')
        if ln > 0:
            msg = msg[:ln]

        txt_level = self._level_map.get(record.levelno, "ERROR")
        self.send_to_logdb(_job_name, txt_level, msg)

    def aggregate_stats(self, msg: str) -> None:
        """Sum stats together, to lessen load on logdb."""

        msg = msg[1:-1]
        for rec in msg.split(", "):
            k, v = rec.split(": ")
            agg = self.stat_cache.get(k, 0)
            if v.find('.') >= 0:
                agg += float(v)
            else:
                agg += int(v)
            self.stat_cache[k] = agg

    def flush_stats(self, service: str) -> None:
        """Send acquired stats to logdb."""
        res = []
        for k, v in self.stat_cache.items():
            res.append("%s: %s" % (k, str(v)))
        if len(res) > 0:
            logmsg = "{%s}" % ", ".join(res)
            self.send_to_logdb(service, "INFO", logmsg)
        self.stat_cache = {}
        self.last_stat_flush = time.time()

    def send_to_logdb(self, service: str, level: str, msg: str) -> None:
        """Actual sending is done here."""

        if self.sock is None:
            self.createSocket()

        if self.sock:
            logcur = self.sock.cursor()
            query = "select * from log.add(%s, %s, %s)"
            logcur.execute(query, [level, service, msg])


# fix unicode bug in SysLogHandler
class SysLogHandler(logging.handlers.SysLogHandler):
    """Fixes unicode bug in logging.handlers.SysLogHandler."""

    # be compatible with both 2.6 and 2.7
    socktype = socket.SOCK_DGRAM

    _udp_reset: float = 0

    def _custom_format(self, record: logging.LogRecord) -> str:
        msg = self.format(record) + '\000'

        # We need to convert record level to lowercase, maybe this will
        # change in the future.
        prio = '<%d>' % self.encodePriority(self.facility,
                                            self.mapPriority(record.levelname))
        msg = prio + msg
        return msg

    def emit(self, record: logging.LogRecord) -> None:
        """
        Emit a record.

        The record is formatted, and then sent to the syslog server. If
        exception information is present, it is NOT sent to the server.
        """
        xmsg = self._custom_format(record)
        msg = xmsg if isinstance(xmsg, bytes) else xmsg.encode("utf8")

        try:
            if self.unixsocket:
                try:
                    self.socket.send(msg)   # type: ignore[has-type]
                except socket.error:
                    self._connect_unixsocket(self.address)  # type: ignore[attr-defined]
                    self.socket.send(msg)                   # type: ignore[has-type]
            elif self.socktype == socket.SOCK_DGRAM:
                now = time.time()
                if now - 1 > self._udp_reset:
                    self.socket.close()  # type: ignore[has-type]
                    self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                    self._udp_reset = now
                self.socket.sendto(msg, self.address)
            else:
                self.socket.sendall(msg)
        except (KeyboardInterrupt, SystemExit):
            raise
        except BaseException:
            self.handleError(record)


class SysLogHostnameHandler(SysLogHandler):
    """Slightly modified standard SysLogHandler - sends also hostname and service type"""

    def _custom_format(self, record: logging.LogRecord) -> str:
        msg = self.format(record)
        format_string = '<%d> %s %s %s\000'
        msg = format_string % (self.encodePriority(self.facility, self.mapPriority(record.levelname)),
                               _hostname, _service_name, msg)
        return msg


# add missing aliases (that are in Logger class)
if not hasattr(LoggerAdapter, 'fatal'):
    LoggerAdapter.fatal = LoggerAdapter.critical    # type: ignore


class SkyLogger(LoggerAdapter):
    """Adds API to existing Logger.
    """
    def trace(self, msg: str, *args: Any, **kwargs: Any) -> None:
        """Log message with severity TRACE."""
        self.log(TRACE, msg, *args, **kwargs)


def getLogger(name: Optional[str] = None, **kwargs_extra: Any) -> SkyLogger:
    """Get logger with extra functionality.

    Adds additional log levels, and extra fields to log record.

    name - name for logging.getLogger()
    kwargs_extra - extra fields to add to log record
    """
    log = logging.getLogger(name)
    return SkyLogger(log, kwargs_extra)




"""Admin scripting.
"""

import inspect
import sys
from typing import Sequence, Optional, Any, Mapping, Callable

import skytools

from .basetypes import Connection, DictRow, ExecuteParams

__all__ = ['AdminScript']


class AdminScript(skytools.DBScript):
    """Contains common admin script tools.

    Second argument (first is .ini file) is taken as command
    name.  If class method 'cmd_' + arg exists, it is called,
    otherwise error is given.
    """
    commands_without_pidfile: Sequence[str] = ()

    def __init__(self, service_name: str, args: Sequence[str]) -> None:
        """AdminScript init."""
        super().__init__(service_name, args)

        if len(self.args) < 2:
            self.log.error("need command")
            sys.exit(1)

        cmd = self.args[1]
        if cmd in self.commands_without_pidfile:
            self.pidfile = None

        if self.pidfile:
            self.pidfile = self.pidfile + ".admin"

    def work(self) -> Optional[int]:
        """Non-looping work function, calls command function."""

        self.set_single_loop(1)

        cmd = self.args[1]
        cmdargs = self.args[2:]

        # find function
        fname = "cmd_" + cmd.replace('-', '_')
        if not hasattr(self, fname):
            self.log.error('bad subcommand, see --help for usage')
            sys.exit(1)
        fn = getattr(self, fname)

        # check if correct number of arguments
        (
                args, varargs, ___varkw, ___defaults,
                ___kwonlyargs, __kwonlydefaults, ___annotations,
        ) = inspect.getfullargspec(fn)
        n_args = len(args) - 1  # drop 'self'
        if varargs is None and n_args != len(cmdargs):
            helpstr = ""
            if n_args:
                helpstr = ": " + " ".join(args[1:])
            self.log.error("command '%s' got %d args, but expects %d%s",
                           cmd, len(cmdargs), n_args, helpstr)
            sys.exit(1)

        # run command
        fn(*cmdargs)

        return None

    def fetch_list(self, db: Connection, sql: str, args: ExecuteParams, keycol: Optional[str] = None) -> Sequence[Any]:
        """Fetch a resultset from db, optionally turning it into value list."""
        curs = db.cursor()
        curs.execute(sql, args)
        rows = curs.fetchall()
        db.commit()
        if not keycol:
            res = rows
        else:
            res = [r[keycol] for r in rows]
        return res

    def display_table(self, db: Connection, desc: str, sql: str, args: ExecuteParams = (), fields: Sequence[str] = (), fieldfmt: Optional[Mapping[str, Callable[[Any], str]]]=None) -> int:
        """Display multirow query as a table."""

        self.log.debug("display_table: %s", skytools.quote_statement(sql, args))
        curs = db.cursor()
        curs.execute(sql, args)
        rows = curs.fetchall()
        db.commit()
        if len(rows) == 0:
            return 0
        if not fieldfmt:
            fieldfmt = {}

        if not fields:
            fields = [f[0] for f in curs.description]

        widths = [15] * len(fields)
        for row in rows:
            for i, k in enumerate(fields):
                rlen = len(str(row[k])) if row[k] else 0
                widths[i] = widths[i] > rlen and widths[i] or rlen
        widths = [w + 2 for w in widths]

        fmt = '%%-%ds' * (len(widths) - 1) + '%%s'
        fmt = fmt % tuple(widths[:-1])
        if desc:
            print(desc)
        print(fmt % tuple(fields))
        print(fmt % tuple('-' * (w - 2) for w in widths))
        #print(fmt % tuple(['-'*15] * len(fields)))
        for row in rows:
            vals = []
            for field in fields:
                val = row[field]
                if field in fieldfmt:
                    val = fieldfmt[field](val)
                vals.append(val)
            print(fmt % tuple(vals))
        print('\n')
        return 1

    def exec_stmt(self, db: Connection, sql: str, args: ExecuteParams) -> None:
        """Run regular non-query SQL on db."""
        self.log.debug("exec_stmt: %s", skytools.quote_statement(sql, args))
        curs = db.cursor()
        curs.execute(sql, args)
        db.commit()

    def exec_query(self, db: Connection, sql: str, args: ExecuteParams) -> Sequence[DictRow]:
        """Run regular query SQL on db."""
        self.log.debug("exec_query: %s", skytools.quote_statement(sql, args))
        curs = db.cursor()
        curs.execute(sql, args)
        res = curs.fetchall()
        db.commit()
        return res




"""Various low-level utility functions for sockets."""


import os
import socket
import sys
from typing import Optional, Union

try:
    import fcntl
except ImportError:
    fcntl = None    # type: ignore

from .basetypes import FileDescriptorLike

__all__ = (
    'set_tcp_keepalive', 'set_nonblocking', 'set_cloexec',
)

SocketLike = Union[socket.socket, FileDescriptorLike]


def set_tcp_keepalive(fd: SocketLike, keepalive: bool = True, tcp_keepidle: int = 4 *
                      60, tcp_keepcnt: int = 4, tcp_keepintvl: int = 15) -> None:
    """Turn on TCP keepalive.  The fd can be either numeric or socket
    object with 'fileno' method.

    OS defaults for SO_KEEPALIVE=1:
     - Linux: (7200, 9, 75) - can configure all.
     - MacOS: (7200, 8, 75) - can configure only tcp_keepidle.
     - Win32: (7200, 5|10, 1) - can configure tcp_keepidle and tcp_keepintvl.

    Our defaults: (240, 4, 15).
    """

    # usable on this OS?
    if not hasattr(socket, 'SO_KEEPALIVE') or not hasattr(socket, 'fromfd'):
        return

    # need socket object
    if isinstance(fd, socket.SocketType):
        s = fd
    else:
        if not isinstance(fd, int):
            fd = fd.fileno()
        s = socket.fromfd(fd, socket.AF_INET, socket.SOCK_STREAM)

    # skip if unix socket
    if getattr(socket, 'AF_UNIX', None):
        if not isinstance(s.getsockname(), tuple):
            return

    # no keepalive?
    if not keepalive:
        s.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 0)
        return

    # basic keepalive
    s.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

    # detect available options
    TCP_KEEPCNT = getattr(socket, 'TCP_KEEPCNT', None)
    TCP_KEEPINTVL = getattr(socket, 'TCP_KEEPINTVL', None)
    TCP_KEEPIDLE = getattr(socket, 'TCP_KEEPIDLE', None)
    TCP_KEEPALIVE = getattr(socket, 'TCP_KEEPALIVE', None)
    SIO_KEEPALIVE_VALS = getattr(socket, 'SIO_KEEPALIVE_VALS', None)
    if TCP_KEEPIDLE is None and TCP_KEEPALIVE is None and sys.platform == 'darwin':
        TCP_KEEPALIVE = 0x10

    # configure
    if TCP_KEEPCNT is not None:
        s.setsockopt(socket.IPPROTO_TCP, TCP_KEEPCNT, tcp_keepcnt)
    if TCP_KEEPINTVL is not None:
        s.setsockopt(socket.IPPROTO_TCP, TCP_KEEPINTVL, tcp_keepintvl)
    if TCP_KEEPIDLE is not None:
        s.setsockopt(socket.IPPROTO_TCP, TCP_KEEPIDLE, tcp_keepidle)
    elif TCP_KEEPALIVE is not None:
        s.setsockopt(socket.IPPROTO_TCP, TCP_KEEPALIVE, tcp_keepidle)
    elif SIO_KEEPALIVE_VALS is not None and fcntl:
        fcntl.ioctl(s.fileno(), SIO_KEEPALIVE_VALS, (1, tcp_keepidle * 1000, tcp_keepintvl * 1000))  # type: ignore


def set_nonblocking(fd: SocketLike, onoff: Optional[bool] = True) -> Optional[bool]:
    """Toggle the O_NONBLOCK flag.

    If onoff==None then return current setting.

    Actual sockets from 'socket' module should use .setblocking() method,
    this is for situations where it is not available.  Eg. pipes
    from 'subprocess' module.
    """
    if fcntl is None:
        return onoff

    flags = fcntl.fcntl(fd, fcntl.F_GETFL)
    if onoff is None:
        return (flags & os.O_NONBLOCK) > 0
    if onoff:
        flags |= os.O_NONBLOCK
    else:
        flags &= ~os.O_NONBLOCK
    fcntl.fcntl(fd, fcntl.F_SETFL, flags)
    return onoff


def set_cloexec(fd: SocketLike, onoff: Optional[bool] = True) -> bool:
    """Toggle the FD_CLOEXEC flag.

    If onoff==None then return current setting.

    Some libraries do it automatically (eg. libpq).
    Others do not (Python stdlib).
    """
    if fcntl is None:
        return True

    flags = fcntl.fcntl(fd, fcntl.F_GETFD)
    if onoff is None:
        return (flags & fcntl.FD_CLOEXEC) > 0
    if onoff:
        flags |= fcntl.FD_CLOEXEC
    else:
        flags &= ~fcntl.FD_CLOEXEC
    fcntl.fcntl(fd, fcntl.F_SETFD, flags)
    return onoff





from typing import Any, Optional, Mapping, Dict

def quote_literal(value: Any) -> str: ...
def quote_copy(value: Any) -> str: ...
def quote_bytea_raw(s: Optional[bytes]) -> Optional[str]: ...
def db_urlencode(dict_val: Mapping[str, Any]) -> str: ...
def db_urldecode(qs: str) -> Dict[str, Optional[str]]: ...
def unescape(val: str) -> str: ...
def unquote_literal(val: str, stdstr: bool = False) -> Optional[str]: ...




"""Check if SQL keywords are up-to-date.
"""

from typing import Dict, List, Tuple

import os.path
import re

import skytools.quoting

versions = [
    "94", "95", "96",
    "10", "11", "12", "13", "14",
    "15", "16", "17", "18", "19"
]

locations = [
    #"/usr/include/postgresql/{VER}/server/parser/kwlist.h",
    "~/src/pgsql/pg{VER}/src/include/parser/kwlist.h",
    #"~/src/pgsql/postgresql/src/include/parser/kwlist.h",
]

known_cats = [
    "COL_NAME_KEYWORD", "UNRESERVED_KEYWORD",
    "RESERVED_KEYWORD", "TYPE_FUNC_NAME_KEYWORD",
]

def addkw(kmap: Dict[str, str], kw: str, cat: str) -> None:
    assert cat in known_cats
    if kw not in kmap:
        kmap[kw] = cat
    elif kmap[kw] != cat and cat not in kmap[kw]:
        kmap[kw] += "," + cat

def _load_kwlist(fn: str, full_map: Dict[str, str], cur_map: Dict[str, str]) -> None:
    fn = os.path.expanduser(fn)
    if not os.path.isfile(fn):
        return
    with open(fn, 'rt') as f:
        data = f.read()
    rc = re.compile(r'PG_KEYWORD[(]"(.*)" , \s* \w+ , \s* (\w+) [)]', re.X)
    for kw, cat in rc.findall(data):
        addkw(full_map, kw, cat)
        if cat == 'UNRESERVED_KEYWORD':
            continue
        #if cat == 'COL_NAME_KEYWORD':
        #    continue
        addkw(cur_map, kw, cat)


def test_kwcheck() -> None:
    """Compare keyword list in quoting.py to the one in postgres sources
    """

    kwset = set(skytools.quoting._ident_kwmap)
    full_map: Dict[str, str] = {}           # all types from kwlist.h
    cur_map: Dict[str, str] = {}            # only kwlist.h
    new_list: List[Tuple[str, str]] = []           # missing from kwset
    obsolete_list: List[Tuple[str, str]] = []      # in kwset, but not in cur_map

    done = set()
    for loc in locations:
        for ver in versions:
            fn = loc.format(VER=ver)
            if fn not in done:
                _load_kwlist(fn, full_map, cur_map)
                done.add(fn)

    if not full_map:
        return

    for kw in sorted(cur_map):
        if kw not in kwset:
            new_list.append((kw, cur_map[kw]))
        kwset.add(kw)

    for k in sorted(kwset):
        if k not in full_map:
            # especially obsolete
            obsolete_list.append((k, '!FULL'))
        elif k not in cur_map:
            # slightly obsolete
            obsolete_list.append((k, '!CUR'))

    assert new_list == []

    # here we need to keep older keywords around longer
    #assert obsolete_list == []




"""Fill gaps in Python time API-s.

parse_iso_timestamp:
    Parse reasonable subset of ISO_8601 timestamp formats.
    [ http://en.wikipedia.org/wiki/ISO_8601 ]

datetime_to_timestamp:
    Get POSIX timestamp from datetime() object.

"""

import re
import time
from datetime import datetime, timedelta, tzinfo
from typing import Optional, Pattern

__all__ = (
    'parse_iso_timestamp', 'FixedOffsetTimezone', 'datetime_to_timestamp',
)


class FixedOffsetTimezone(tzinfo):
    """Fixed offset in minutes east from UTC."""
    __slots__ = ('__offset', '__name')

    __offset: timedelta
    __name: str

    def __init__(self, offset: int) -> None:
        super().__init__()

        self.__offset = timedelta(minutes=offset)

        # numeric tz name
        h, m = divmod(abs(offset), 60)
        if offset < 0:
            h = -h
        if m:
            self.__name = "%+03d:%02d" % (h, m)
        else:
            self.__name = "%+03d" % h

    def utcoffset(self, dt: Optional[datetime]) -> Optional[timedelta]:
        return self.__offset

    def tzname(self, dt: Optional[datetime]) -> Optional[str]:
        return self.__name

    def dst(self, dt: Optional[datetime]) -> Optional[timedelta]:
        return ZERO


ZERO = timedelta(0)

#
# Parse ISO_8601 timestamps.
#

"""
TODO:
- support more combinations from ISO 8601 (only reasonable ones)
- cache TZ objects
- make it faster?
"""

_iso_regex = r"""
    \s*
    (?P<year> \d\d\d\d) [-] (?P<month> \d\d) [-] (?P<day> \d\d) [ T]
    (?P<hour> \d\d) [:] (?P<min> \d\d)
    (?: [:] (?P<sec> \d\d ) (?: [.,] (?P<ss> \d+))? )?
    (?: \s*  (?P<tzsign> [-+]) (?P<tzhr> \d\d) (?: [:]? (?P<tzmin> \d\d))?
      | (?P<tzname> Z ) )?
    \s* $
    """
_iso_rc: Optional[Pattern[str]] = None


def parse_iso_timestamp(s: str, default_tz: Optional[tzinfo] = None) -> datetime:
    """Parse ISO timestamp to datetime object.

    YYYY-MM-DD[ T]HH:MM[:SS[.ss]][-+HH[:MM]]

    Assumes that second fractions are zero-trimmed from the end,
    so '.15' means 150000 microseconds.

    If the timezone offset is not present, use default_tz as tzinfo.
    By default its None, meaning the datetime object will be without tz.

    Only fixed offset timezones are supported.
    """

    global _iso_rc
    if _iso_rc is None:
        _iso_rc = re.compile(_iso_regex, re.X)

    m = _iso_rc.match(s)
    if not m:
        raise ValueError('Date not in ISO format: %s' % repr(s))

    tz = default_tz
    if m.group('tzsign'):
        tzofs = int(m.group('tzhr')) * 60
        if m.group('tzmin'):
            tzofs += int(m.group('tzmin'))
        if m.group('tzsign') == '-':
            tzofs = -tzofs
        tz = FixedOffsetTimezone(tzofs)
    elif m.group('tzname'):
        tz = UTC

    return datetime(
        int(m.group('year')), int(m.group('month')), int(m.group('day')),
        int(m.group('hour')), int(m.group('min')),
        m.group('sec') and int(m.group('sec')) or 0,
        m.group('ss') and int(m.group('ss').ljust(6, '0')) or 0,
        tz
    )


#
# POSIX timestamp from datetime()
#

UTC = FixedOffsetTimezone(0)
TZ_EPOCH = datetime.fromtimestamp(0, UTC)
UTC_NOTZ_EPOCH = datetime.utcfromtimestamp(0)


def datetime_to_timestamp(dt: datetime, local_time: bool = True) -> float:
    """Get posix timestamp from datetime() object.

    if dt is without timezone, then local_time specifies
    whether it's UTC or local time.

    Returns seconds since epoch as float.
    """
    if dt.tzinfo:
        delta = dt - TZ_EPOCH
        return delta.total_seconds()
    elif local_time:
        s = time.mktime(dt.timetuple())
        return s + (dt.microsecond / 1000000.0)
    else:
        delta = dt - UTC_NOTZ_EPOCH
        return delta.total_seconds()





import os
import tempfile

from skytools.fileutil import write_atomic


def test_write_atomic() -> None:
    pidfn = tempfile.mktemp('.pid')
    write_atomic(pidfn, "1")
    write_atomic(pidfn, "2")
    os.remove(pidfn)


def test_write_atomic_bak() -> None:
    pidfn = tempfile.mktemp('.pid')
    write_atomic(pidfn, "1", '.bak')
    write_atomic(pidfn, "2", '.bak')
    os.remove(pidfn)




""" Class used to handle multiset receiving and returning PL/Python procedures
"""

import logging
from typing import List, Optional, Sequence, Any, Dict, Union, Tuple

import skytools
from skytools import dbdict

try:
    import plpy
except ImportError:
    pass

__all__ = (
    'DBService', 'ServiceContext',
    'get_record', 'get_record_list',
    'make_record', 'make_record_array',
    'TableAPI',
    #'log_result', 'transform_fields'
)


def transform_fields(rows: Sequence[Dict[str, Any]], key_fields: Sequence[str], name_field: str,
                     data_field: str) -> List[Dict[str, Any]]:
    """Convert multiple-rows per key input array
    to one-row, multiple-column output array.  The input arrays
    must be sorted by the key fields.
    """
    cur_key: List[str] = []
    cur_row: Dict[str, Any] = {}
    res = []
    for r in rows:
        k = [r[f] for f in key_fields]
        if k != cur_key:
            cur_key = k
            cur_row = {}
            for f in key_fields:
                cur_row[f] = r[f]
            res.append(cur_row)
        cur_row[r[name_field]] = r[data_field]
    return res


# render_table

def render_table(rows: Sequence[Dict[str, Any]], fields: Sequence[str]) -> List[str]:
    """ Render result rows as a table.
        Returns array of lines.
    """
    widths = [15] * len(fields)
    for row in rows:
        for i, k in enumerate(fields):
            rlen = len(str(row.get(k)))
            widths[i] = widths[i] > rlen and widths[i] or rlen
    widths = [w + 2 for w in widths]

    fmt = '%%-%ds' * (len(widths) - 1) + '%%s'
    fmt = fmt % tuple(widths[:-1])

    lines = []
    lines.append(fmt % tuple(fields))
    lines.append(fmt % tuple(['-' * 15] * len(fields)))
    for row in rows:
        lines.append(fmt % tuple(str(row.get(k)) for k in fields))
    return lines


# data conversion to and from url

def get_record(arg: Optional[str]) -> dbdict:
    """ Parse data for one urlencoded record.
        Useful for turning incoming serialized data into structure usable for manipulation.
    """
    if not arg:
        return dbdict()

    # allow array of single record
    if arg[0] in ('{', '['):
        lst = skytools.parse_pgarray(arg)
        if not lst or len(lst) != 1:
            raise ValueError('get_record() expects exactly 1 row, got %d' % len(lst or []))
        arg = lst[0]
        if not arg:
            return dbdict()

    # parse record
    return dbdict(skytools.db_urldecode(arg))


def get_record_list(array: Optional[Union[str, List[Optional[str]]]]) -> List[dbdict]:
    """ Parse array of urlencoded records.
        Useful for turning incoming serialized data into structure usable for manipulation.
    """
    if array is None:
        return []

    if not isinstance(array, list):
        array = skytools.parse_pgarray(array) or []
    return [get_record(el) for el in array if el is not None]


def get_record_lists(tbl: Sequence[Dict[str, Any]], field: str) -> dbdict:
    """ Create dictionary of lists from given list using field as grouping criteria
        Used for master detail operatons to group detail records according to master id
    """
    records = dbdict()
    for rec in tbl:
        master_id = str(rec[field])
        records.setdefault(master_id, []).append(rec)
    return records


def _make_record_convert(row: Dict[str, Any]) -> str:
    """Converts complex values."""
    d = row.copy()
    for k, v in d.items():
        if isinstance(v, list):
            d[k] = skytools.make_pgarray(v)
    return skytools.db_urlencode(d)


def make_record(row: Dict[str, Any]) -> str:
    """ Takes record as dict and returns it as urlencoded string.
        Used to send data out of db service layer.or to fake incoming calls
    """
    for v in row.values():
        if isinstance(v, list):
            return _make_record_convert(row)
    return skytools.db_urlencode(row)


def make_record_array(rowlist: Sequence[Dict[str, Any]]):
    """ Takes list of records got from plpy execute and turns it into postgers aray string.
        Used to send data out of db service layer.
    """
    return '{' + ','.join([make_record(row) for row in rowlist]) + '}'


def get_result_items(rec_list: Sequence[Dict[str, Any]], name: str) -> Optional[List[dbdict]]:
    """ Get return values from result
    """
    for r in rec_list:
        if r['res_code'] == name:
            return get_record_list(r['res_rows'])
    return None


def log_result(log: logging.Logger, rec_list: Sequence[Dict[str, Any]]) -> None:
    """ Sends dbservice execution logs to logfile
    """
    msglist = get_result_items(rec_list, "_status")
    if msglist is None:
        if rec_list:
            log.warning('Unhandled output result: _status res_code not present.')
    else:
        for msg in msglist:
            log.debug(msg['_message'])


class DBService:
    """  Wrap parameterized query handling and multiset stored procedure writing
    """
    ROW = "_row"            # name of the fake field where internal record id is stored
    FIELD = "_field"        # parameter name for the field in record that is related to current message
    PARAM = "_param"        # name of the parameter to which message relates
    SKIP = "skip"           # used when record is needed for it's data but is not been updated
    INSERT = "insert"
    UPDATE = "update"
    DELETE = "delete"
    INFO = "info"           # just informative message for the user
    NOTICE = "notice"       # more than info less than warning
    WARNING = "warning"     # warning message, something is out of ordinary
    ERROR = "error"         # error found but execution continues until check then error is raised
    FATAL = "fatal"         # execution is terminated at once and all found errors returned

    rows_found: int = 0

    _context: str
    _is_test: bool
    _retval: List[Tuple[str, List[dbdict]]]
    global_dict: Optional[Dict[str, Any]]
    sqls: Optional[List[Dict[str, str]]]
    can_save: bool
    messages: List[Dict[str, Any]]

    def __init__(self, context: str, global_dict: Optional[Dict[str, Any]] = None) -> None:
        """ This object must be initiated in the beginning of each db service
        """
        rec = skytools.db_urldecode(context)
        self._context = context             # used to run dbservice in retval
        self.global_dict = global_dict      # used for cacheing query plans
        self._retval = []                   # used to collect return resultsets
        self._is_test = 'is_test' in rec    # used to convert output into human readable form

        self.sqls = None                    # if sqls stays None then no recording of sqls is done
        if "show_sql" in rec:               # api must add exected sql to resultset
            self.sqls = []                  # sql's executed by dbservice, used for dubugging

        self.can_save = True                # used to keep value most severe error found so far
        self.messages = []                  # used to hold list of messages to be returned to the user

    # error and message handling

    def tell_user(self, severity: str, code: str, message: str, params: Optional[Dict[str, Any]] = None, **kvargs: Any) -> None:
        """ Adds another message to the set of messages to be sent back to user
            If error message then can_save is set false
            If fatal message then error or found errors are raised at once
        """
        params = params or kvargs
        #plpy.notice("%s %s: %s %s" % (severity, code, message, str(params)))
        params["_severity"] = severity
        params["_code"] = code
        params["_message"] = message
        self.messages.append(params)
        if severity == self.ERROR:
            self.can_save = False
        if severity == self.FATAL:
            self.can_save = False
            self.raise_if_errors()

    def raise_if_errors(self) -> None:
        """ To be used in places where before continuing must be chcked if errors have been found
            Raises found errors packing them into error message as urlencoded string
        """
        if not self.can_save:
            msgs = "Dbservice error(s): " + make_record_array(self.messages)
            plpy.error(msgs)

    # run sql meant mostly for select but not limited to

    def create_query(self, sql: str, params: Optional[Dict[str, Any]] = None, **kvargs: Any) -> skytools.PLPyQueryBuilder:
        """ Returns initialized querybuilder object for building complex dynamic queries
        """
        params = params or kvargs
        return skytools.PLPyQueryBuilder(sql, params, self.global_dict, self.sqls)

    def run_query(self, sql: str, params: Optional[Dict[str, Any]] = None, **kvargs: Any) -> List[dbdict]:
        """ Helper function if everything you need is just paramertisized execute
            Sets rows_found that is coneninet to use when you don't need result just
            want to know how many rows were affected
        """
        params = params or kvargs
        rows = skytools.plpy_exec(self.global_dict, sql, params)
        # convert result rows to dbdict
        if rows:
            rows = [dbdict(r) for r in rows]
            self.rows_found = len(rows)
        else:
            self.rows_found = 0
            rows = []
        return rows

    def run_query_row(self, sql: str, params: Optional[Dict[str, Any]] = None, **kvargs: Any) -> Optional[dbdict]:
        """ Helper function if everything you need is just paramertisized execute to
            fetch one row only. If not found none is returned
        """
        params = params or kvargs
        rows = self.run_query(sql, params)
        if len(rows) == 0:
            return None
        return rows[0]

    def run_exists(self, sql: str, params: Optional[Dict[str, Any]] = None, **kvargs: Any) -> int:
        """ Helper function to find out that record in given table exists using
            values in dict as criteria. Takes away all the hassle of preparing statements
            and processing returned result giving out just one boolean
        """
        params = params or kvargs
        self.run_query(sql, params)
        return self.rows_found

    def run_lookup(self, sql: str, params: Optional[Dict[str, Any]] = None, **kvargs: Any):
        """ Helper function to fetch one value Takes away all the hassle of preparing statements
            and processing returned result giving out just one value. Uses plan cache if used inside
            db service
        """
        params = params or kvargs
        rows = self.run_query(sql, params)
        if len(rows) == 0:
            return None
        row = rows[0]
        return list(row.values())[0]

    # resultset handling

    def return_next(self, rows: List[dbdict], res_name: str, severity: Optional[str] = None) -> List[dbdict]:
        """ Adds given set of rows to resultset
        """
        self._retval.append((res_name, rows))
        if severity is not None and len(rows) == 0:
            self.tell_user(severity, "dbsXXXX", "No matching records found")
        return rows

    def return_next_sql(self, sql: str, params: Optional[Dict[str, Any]], res_name: str, severity: Optional[str] = None) -> List[dbdict]:
        """ Exectes query and adds recors resultset
        """
        rows = self.run_query(sql, params)
        return self.return_next(rows, res_name, severity)

    def retval(self, service_name: Optional[str] = None, params: Optional[Dict[str, Any]] = None, **kvargs: Any) -> List[Tuple[str, str, str]]:
        """ Return collected resultsets and append to the end messages to the users
            Method is called usually as last statement in dbservice to return the results
            Also converts results into desired format
        """
        params = params or kvargs
        self.raise_if_errors()
        if len(self.messages):
            self.return_next(self.messages, "_status")  # type: ignore
        if self.sqls is not None and len(self.sqls):
            self.return_next(self.sqls, "_sql") # type: ignore
        results: List[Tuple[str, str, str]] = []
        for r in self._retval:
            res_name = r[0]
            rows = r[1]
            res_count = str(len(rows))
            if self._is_test and len(rows) > 0:
                results.append((res_name, res_count, res_name))
                n = 1
                for trow in render_table(rows, list(rows[0].keys())):
                    results.append((res_name, str(n), trow))
                    n += 1
            else:
                res_rows = make_record_array(rows)
                results.append((res_name, res_count, res_rows))
        if service_name:
            sql = "select * from %s( {i_context}, {i_params} );" % skytools.quote_fqident(service_name)
            par = dbdict(i_context=self._context, i_params=make_record(params))
            res = self.run_query(sql, par)
            for row in res:
                results.append((row.res_code, row.res_text, row.res_rows))
        return results

    # miscellaneous

    def check_required(self, record_name, record, severity, *fields):
        """ Checks if all required fields are present in record
            Used to validate incoming data
            Returns list of field names that are missing or empty
        """
        missing = []
        params = {self.PARAM: record_name}
        if self.ROW in record:
            params[self.ROW] = record[self.ROW]
        for field in fields:
            params[self.FIELD] = field
            if field in record:
                if record[field] is None or (isinstance(record[field], str) and len(record[field]) == 0):
                    self.tell_user(severity, "dbsXXXX", "Required value missing: {%s}.{%s}" % (
                                   self.PARAM, self.FIELD), **params)
                    missing.append(field)
            else:
                self.tell_user(severity, "dbsXXXX", "Required field missing: {%s}.{%s}" % (
                               self.PARAM, self.FIELD), **params)
                missing.append(field)
        return missing


# TableAPI
class TableAPI:
    """ Class for managing one record updates using primary key
    """
    _table = None   # schema name and table name
    _where = None   # where condition used for update and delete
    _id = None      # name of the primary key filed
    _id_type = None  # column type of primary key
    _op = None      # operation currently carried out
    _ctx = None     # context object for username and version
    _logging = True  # should tapi log data changed
    _row = None     # row identifer from calling program

    def __init__(self, ctx, table, create_log=True, id_type='int8'):
        """ Table name is used to construct insert update and delete statements
            Table must have primary key field whose name is in format id_<table>
            Tablename should be in format schema.tablename
        """
        self._ctx = ctx
        self._table = skytools.quote_fqident(table)
        self._id = "id_" + skytools.fq_name_parts(table)[1]
        self._id_type = id_type
        self._where = '%s = {%s:%s}' % (skytools.quote_ident(self._id), self._id, self._id_type)
        self._logging = create_log

    def _log(self, result, original=None):
        """ Log changei into table log.changelog
        """
        if not self._logging:
            return
        assert self._ctx
        changes = []
        for key in result.keys():
            if self._op == 'update':
                if key in original:
                    if str(original[key]) != str(result[key]):
                        changes.append(key + ": " + str(original[key]) + " -> " + str(result[key]))
            else:
                changes.append(key + ": " + str(result[key]))
        self._ctx.log(self._table, result[self._id], self._op, "\n".join(changes))

    def _version_check(self, original, version):
        assert self._ctx
        if version is None:
            self._ctx.tell_user(
                self._ctx.INFO, "dbsXXXX",
                "Record ({table}.{field}={id}) has been deleted by other user "
                "while you were editing. Check version ({ver}) in changelog for details.",
                table=self._table, field=self._id, id=original[self._id],
                ver=original.version, _row=self._row
            )
        if version is not None and original.version is not None:
            if int(version) != int(original.version) and self._ctx:
                self._ctx.tell_user(
                    self._ctx.INFO, "dbsXXXX",
                    "Record ({table}.{field}={id}) has been changed by other user while you were editing. "
                    "Version in db: ({db_ver}) and version sent by caller ({caller_ver}). "
                    "See changelog for details.",
                    table=self._table, field=self._id, id=original[self._id], db_ver=original.version,
                    caller_ver=version, _row=self._row
                )

    def _insert(self, data):
        assert self._ctx
        fields = []
        values = []
        for key in data.keys():
            if data[key] is not None:       # ignore empty
                fields.append(skytools.quote_ident(key))
                values.append("{" + key + "}")
        sql = "insert into %s (%s) values (%s) returning *;" % (
            self._table, ",".join(fields), ",".join(values)
        )
        result = self._ctx.run_query_row(sql, data)
        self._log(result)
        return result

    def _update(self, data, version):
        assert self._ctx
        sql = "select * from %s where %s" % (self._table, self._where)
        original = self._ctx.run_query_row(sql, data)
        self._version_check(original, version)
        pairs = []
        for key in data.keys():
            if data[key] is None:
                pairs.append(key + " = NULL")
            else:
                pairs.append(key + " = {" + key + "}")
        sql = "update %s set %s where %s returning *;" % (self._table, ", ".join(pairs), self._where)
        assert self._ctx
        result = self._ctx.run_query_row(sql, data)
        self._log(result, original)
        return result

    def _delete(self, data, version):
        assert self._ctx
        sql = "delete from %s where %s returning *;" % (self._table, self._where)
        result = self._ctx.run_query_row(sql, data)
        self._version_check(result, version)
        self._log(result)
        return result

    def do(self, data):
        """ Do dml according to special field _op that must be given together wit data
        """
        assert self._ctx
        result = data                               # so it is initialized for skip
        self._op = data.pop(self._ctx.OP)           # determines operation done
        self._row = data.pop(self._ctx.ROW, None)   # internal record id used for error reporting
        if self._row is None:                       # if no _row variable was provided
            self._row = data.get(self._id, None)    # use id instead
        if self._id in data and data[self._id]:     # if _id field is given
            if int(data[self._id]) < 0:             # and it is fake key generated by ui
                data.pop(self._id)                  # remove fake key so real one can be assigned
        version = data.get('version', None)         # version sent from caller
        data['version'] = self._ctx.version         # current transaction id is stored in each record
        if self._op == self._ctx.INSERT:
            result = self._insert(data)
        elif self._op == self._ctx.UPDATE:
            result = self._update(data, version)
        elif self._op == self._ctx.DELETE:
            result = self._delete(data, version)
        elif self._op == self._ctx.SKIP:
            pass
        elif self._ctx:
            self._ctx.tell_user(self._ctx.ERROR, "dbsXXXX",
                                "Unahndled _op='{op}' value in TableAPI (table={table}, id={id})",
                                op=self._op, table=self._table, id=data[self._id])
        if self._ctx:
            result[self._ctx.OP] = self._op
            result[self._ctx.ROW] = self._row
        return result


# ServiceContext

class ServiceContext(DBService):
    OP = "_op"              # name of the fake field where record modificaton operation is stored

    def __init__(self, context: str, global_dict: Optional[Dict[str, Any]] = None) -> None:
        """ This object must be initiated in the beginning of each db service
        """
        super().__init__(context, global_dict)

        rec = skytools.db_urldecode(context)
        if "username" not in rec:
            plpy.error("Username must be provided in db service context parameter")
        self.username = rec['username']     # used for logging purposes

        res = plpy.execute("select txid_current() as txid;")
        row = res[0]
        self.version = row["txid"]
        self.rows_found = 0                 # Flag set by run query to inicate number of rows got

    # logging

    def log(self, _object_type, _key_object, _change_op, _payload):
        """ Log stuff into the changelog whatever seems relevant to be logged
        """
        self.run_query(
            "select log.log_change( {version}, {username}, {object_type}, {key_object}, {change_op}, {payload} );",
            version=self.version, username=self.username,
            object_type=_object_type, key_object=_key_object,
            change_op=_change_op, payload=_payload)

    # data conversion to and from url

    def get_record(self, arg):
        """ Parse data for one urlencoded record.
            Useful for turning incoming serialized data into structure usable for manipulation.
        """
        return get_record(arg)

    def get_record_list(self, array):
        """ Parse array of urlencoded records.
            Useful for turning incoming serialized data into structure usable for manipulation.
        """
        return get_record_list(array)

    def get_list_groups(self, tbl, field):
        """ Create dictionary of lists from given list using field as grouping criteria
            Used for master detail operatons to group detail records according to master id
        """
        return get_record_lists(tbl, field)

    def make_record(self, row):
        """ Takes record as dict and returns it as urlencoded string.
            Used to send data out of db service layer.or to fake incoming calls
        """
        return make_record(row)

    def make_record_array(self, rowlist):
        """ Takes list of records got from plpy execute and turns it into postgers aray string.
            Used to send data out of db service layer.
        """
        return make_record_array(rowlist)

    # tapi based dml functions

    def _changelog(self, fields):
        log = True
        if fields:
            if '_log' in fields:
                if not fields.pop('_log'):
                    log = False
            if '_log_id' in fields:
                fields.pop('_log_id')
            if '_log_field' in fields:
                fields.pop('_log_field')
        return log

    def tapi_do(self, tablename, row, **fields):
        """ Convenience function for just doing the change without creating tapi object first
            Fields object may contain aditional overriding values that are applied before do
        """
        tapi = TableAPI(self, tablename, self._changelog(fields))
        row = row or dbdict()
        if fields:
            row.update(fields)
        return tapi.do(row)

    def tapi_do_set(self, tablename, rows, **fields):
        """ Does changes to list of detail rows
            Used for normal foreign keys in master detail relationships
            Dows first deletes then updates and then inserts to avoid uniqueness problems
        """
        tapi = TableAPI(self, tablename, self._changelog(fields))
        results, updates, inserts = [], [], []
        for row in rows:
            if fields:
                row.update(fields)
            if row[self.OP] == self.DELETE:
                results.append(tapi.do(row))
            elif row[self.OP] == self.UPDATE:
                updates.append(row)
            else:
                inserts.append(row)
        for row in updates:
            results.append(tapi.do(row))
        for row in inserts:
            results.append(tapi.do(row))
        return results

    # resultset handling

    def retval_dbservice(self, service_name: str, ctx: str, **params: Any) -> List[Tuple[str, str, str]]:
        """ Runs service with standard interface.
            Convenient to use for calling select services from other services
            For example to return data after doing save
        """
        self.raise_if_errors()
        service_sql = "select * from %s( {i_context}, {i_params} );" % skytools.quote_fqident(service_name)
        service_params = {"i_context": ctx, "i_params": self.make_record(params)}
        results = self.run_query(service_sql, service_params)
        retval = self.retval()
        for r in results:
            retval.append((r.res_code, r.res_text, r.res_rows))
        return retval

    # miscellaneous

    def field_copy(self, rec, *keys):
        """ Used to copy subset of fields from one record into another
            example: dbs.copy(record, hosting) "start_date", "key_colo", "key_rack")
        """
        retval = dbdict()
        for key in keys:
            if key in rec:
                retval[key] = rec[key]
        return retval

    def field_set(self, **fields):
        """ Fills dict with given values and returns resulting dict
            If dict was not provied with call it is created
        """
        return fields





from skytools.natsort import (
    natsort, natsort_icase, natsort_key, natsorted, natsorted_icase,
)


def test_natsorted() -> None:
    res = natsorted(['1', 'ver-1.11', '', 'ver-1.0'])
    assert res == ['', '1', 'ver-1.0', 'ver-1.11']


def test_natsort() -> None:
    res = ['a1', '2a', '.1']
    natsort(res)
    assert res == ['.1', '2a', 'a1']


def test_natsorted_icase() -> None:
    res = natsorted_icase(['Ver-1.1', 'vEr-1.11', '', 'veR-1.0'])
    assert res == ['', 'veR-1.0', 'Ver-1.1', 'vEr-1.11']


def test_natsort_icase() -> None:
    res = ['Ver-1.1', 'vEr-1.11', '', 'veR-1.0']
    natsort_icase(res)
    assert res == ['', 'veR-1.0', 'Ver-1.1', 'vEr-1.11']


def _natcmp(a: str, b: str) -> str:
    k1 = natsort_key(a)
    k2 = natsort_key(b)
    if k1 < k2:
        return 'ok'
    return f"fail: a='{a}' > b='{b}'"


def test_natsort_order() -> None:
    assert _natcmp('1', '2') == 'ok'
    assert _natcmp('2', '11') == 'ok'
    assert _natcmp('.', '1') == 'ok'
    assert _natcmp('1', 'a') == 'ok'
    assert _natcmp('a~1', 'ab') == 'ok'
    assert _natcmp('a~1', 'a') == 'ok'
    assert _natcmp('a~1', 'a1') == 'ok'
    assert _natcmp('00', '0') == 'ok'
    assert _natcmp('001', '0') == 'ok'
    assert _natcmp('0', '01') == 'ok'
    assert _natcmp('011', '02') == 'ok'
    assert _natcmp('00~1', '0~1') == 'ok'
    assert _natcmp('~~~', '~~') == 'ok'
    assert _natcmp('1~beta0', '1') == 'ok'
    assert _natcmp('1', '1.0') == 'ok'

    assert _natcmp('~', '') == 'ok'
    assert _natcmp('', '0') == 'ok'
    assert _natcmp('', '1') == 'ok'
    assert _natcmp('', 'a') == 'ok'




"""Extra tests for quoting module.
"""

from typing import Dict, Callable, Any, List, Type, Tuple, Union, Optional
from decimal import Decimal

try:
    import psycopg2.extras
    _have_psycopg2 = True
except ImportError:
    _have_psycopg2 = False

import pytest

import skytools._cquoting
import skytools._pyquoting
from skytools.quoting import (
    json_decode, json_encode, make_pgarray,
    quote_fqident, unescape_copy, unquote_fqident,
)

QuoteTests = List[Tuple[Any, str]]
QuoteFunc = Callable[[Any], str]
UnquoteTests = List[Tuple[str, Optional[str]]]
UnquoteFunc = Callable[[str], Optional[str]]

class fake_cursor:
    """create a DictCursor row"""
    index = {'id': 0, 'data': 1}
    description = ['x', 'x']


if _have_psycopg2:
    dbrow = psycopg2.extras.DictRow(fake_cursor())
    dbrow[0] = '123'
    dbrow[1] = 'value'


def try_quote(func: QuoteFunc, data_list: QuoteTests) -> None:
    for val, exp in data_list:
        got = func(val)
        assert got == exp


def try_unquote(func: UnquoteFunc, data_list: UnquoteTests) -> None:
    for val, exp in data_list:
        got = func(val)
        assert got == exp


def test_quote_literal() -> None:
    sql_literal = [
        (None, "null"),
        ("", "''"),
        ("a'b", "'a''b'"),
        (r"a\'b", r"E'a\\''b'"),
        (1, "'1'"),
        (True, "'True'"),
        (Decimal(1), "'1'"),
        (u'qwe', "'qwe'")
    ]
    def quote_literal_c(x: Any) -> str:
        return skytools._cquoting.quote_literal(x)
    def quote_literal_py(x: Any) -> str:
        return skytools._pyquoting.quote_literal(x)
    def quote_literal_default(x: Any) -> str:
        return skytools.quote_literal(x)
    try_quote(quote_literal_default, sql_literal)
    try_quote(quote_literal_default, sql_literal)
    try_quote(quote_literal_default, sql_literal)


qliterals_common = [
    (r"""null""", None),
    (r"""NULL""", None),
    (r"""123""", "123"),
    (r"""''""", r""""""),
    (r"""'a''b''c'""", r"""a'b'c"""),
    (r"""'foo'""", r"""foo"""),
    (r"""E'foo'""", r"""foo"""),
    (r"""E'a\n\t\a\b\0\z\'b'""", "a\n\t\x07\x08\x00z'b"),
    (r"""$$$$""", r""),
    (r"""$$qw$e$z$$""", r"qw$e$z"),
    (r"""$qq$$aa$$$'"\\$qq$""", '$aa$$$\'"\\\\'),
    (u"'qwe'", 'qwe'),
]

bad_dol_literals = [
    ('$$', '$$'),
    #('$$q', '$$q'),
    ('$$q$', '$$q$'),
    ('$q$q$', '$q$q$'),
    ('$q$q$x$', '$q$q$x$'),
]


def test_unquote_literal() -> None:
    qliterals_nonstd = qliterals_common + [
        (r"""'a\\b\\c'""", r"""a\b\c"""),
        (r"""e'a\\b\\c'""", r"""a\b\c"""),
    ]
    def unquote_literal_c(x: str) -> Optional[str]:
        return skytools._cquoting.unquote_literal(x)
    def unquote_literal_py(x: str) -> Optional[str]:
        return skytools._pyquoting.unquote_literal(x)
    def unquote_literal_default(x: str) -> Optional[str]:
        return skytools.unquote_literal(x)
    try_unquote(unquote_literal_c, qliterals_nonstd)
    try_unquote(unquote_literal_py, qliterals_nonstd)
    try_unquote(unquote_literal_default, qliterals_nonstd)

    for v1, v2 in bad_dol_literals:
        with pytest.raises(ValueError):
            skytools._pyquoting.unquote_literal(v1)
        with pytest.raises(ValueError):
            skytools._cquoting.unquote_literal(v1)
        with pytest.raises(ValueError):
            skytools.unquote_literal(v1)


def test_unquote_literal_std() -> None:
    qliterals_std = qliterals_common + [
        (r"''", r""),
        (r"'foo'", r"foo"),
        (r"E'foo'", r"foo"),
        (r"'\\''z'", r"\\'z"),
    ]
    for val, exp in qliterals_std:
        assert skytools._cquoting.unquote_literal(val, True) == exp
        assert skytools._pyquoting.unquote_literal(val, True) == exp
        assert skytools.unquote_literal(val, True) == exp


def test_quote_copy() -> None:
    sql_copy = [
        (None, "\\N"),
        ("", ""),
        ("a'\tb", "a'\\tb"),
        (r"a\'b", r"a\\'b"),
        (1, "1"),
        (True, "True"),
        (u"qwe", "qwe"),
        (Decimal(1), "1"),
    ]
    try_quote(skytools._cquoting.quote_copy, sql_copy)
    try_quote(skytools._pyquoting.quote_copy, sql_copy)
    try_quote(skytools.quote_copy, sql_copy)


def test_quote_bytea_raw() -> None:
    sql_bytea_raw: List[Tuple[Optional[bytes], Optional[str]]] = [
        (None, None),
        (b"", ""),
        (b"a'\tb", "a'\\011b"),
        (b"a\\'b", r"a\\'b"),
        (b"\t\344", r"\011\344"),
    ]
    for val, exp in sql_bytea_raw:
        assert skytools._cquoting.quote_bytea_raw(val) == exp
        assert skytools._pyquoting.quote_bytea_raw(val) == exp
        assert skytools.quote_bytea_raw(val) == exp


def test_quote_bytea_raw_fail() -> None:
    with pytest.raises(TypeError):
        skytools._pyquoting.quote_bytea_raw(u'qwe')  # type: ignore[arg-type]
    #assert_raises(TypeError, skytools._cquoting.quote_bytea_raw, u'qwe')
    #assert_raises(TypeError, skytools.quote_bytea_raw, 'qwe')


def test_quote_ident() -> None:
    sql_ident = [
        ('', '""'),
        ("a'\t\\\"b", '"a\'\t\\""b"'),
        ('abc_19', 'abc_19'),
        ('from', '"from"'),
        ('0foo', '"0foo"'),
        ('mixCase', '"mixCase"'),
        (u'utf', 'utf'),
    ]
    try_quote(skytools.quote_ident, sql_ident)


def test_fqident() -> None:
    assert quote_fqident('tbl') == 'public.tbl'
    assert quote_fqident('Baz.Foo.Bar') == '"Baz"."Foo.Bar"'


def _sort_urlenc(func: Callable[[Any], str]) -> Callable[[Any], str]:
    def wrapper(data: Any) -> str:
        res = func(data)
        return '&'.join(sorted(res.split('&')))
    return wrapper


def test_db_urlencode() -> None:
    t_urlenc = [
        ({}, ""),
        ({'a': 1}, "a=1"),
        ({'a': None}, "a"),
        ({'qwe': 1, u'zz': u"qwe"}, 'qwe=1&zz=qwe'),
        ({'qwe': 1, u'zz': u"qwe"}, 'qwe=1&zz=qwe'),
        ({'a': '\000%&'}, "a=%00%25%26"),
        ({'a': Decimal("1")}, "a=1"),
    ]
    if _have_psycopg2:
        t_urlenc.append((dbrow, 'data=value&id=123'))
    try_quote(_sort_urlenc(skytools._cquoting.db_urlencode), t_urlenc)
    try_quote(_sort_urlenc(skytools._pyquoting.db_urlencode), t_urlenc)
    try_quote(_sort_urlenc(skytools.db_urlencode), t_urlenc)


def test_db_urldecode() -> None:
    t_urldec = [
        ("", {}),
        ("a=b&c", {'a': 'b', 'c': None}),
        ("&&b=f&&", {'b': 'f'}),
        (u"abc=qwe", {'abc': 'qwe'}),
        ("b=", {'b': ''}),
        ("b=%00%45", {'b': '\x00E'}),
    ]
    for val, exp in t_urldec:
        assert skytools._cquoting.db_urldecode(val) == exp
        assert skytools._pyquoting.db_urldecode(val) == exp
        assert skytools.db_urldecode(val) == exp


def test_unescape() -> None:
    t_unesc: List[Tuple[str, str]] = [
        ("", ""),
        ("\\N", "N"),
        ("abc", "abc"),
        (r"\0\000\001\01\1", "\0\000\001\001\001"),
        (r"a\001b\tc\r\n", "a\001b\tc\r\n"),
    ]
    for val, exp in t_unesc:
        assert skytools._cquoting.unescape(val) == exp
        assert skytools._pyquoting.unescape(val) == exp
        assert skytools.unescape(val) == exp


def test_quote_bytea_literal() -> None:
    bytea_raw = [
        (None, "null"),
        (b"", "''"),
        (b"a'\tb", "E'a''\\\\011b'"),
        (b"a\\'b", r"E'a\\\\''b'"),
        (b"\t\344", r"E'\\011\\344'"),
    ]
    try_quote(skytools.quote_bytea_literal, bytea_raw)


def test_quote_bytea_copy() -> None:
    bytea_raw = [
        (None, "\\N"),
        (b"", ""),
        (b"a'\tb", "a'\\\\011b"),
        (b"a\\'b", r"a\\\\'b"),
        (b"\t\344", r"\\011\\344"),
    ]
    try_quote(skytools.quote_bytea_copy, bytea_raw)


def test_quote_statement() -> None:
    sql = "set a=%s, b=%s, c=%s"
    args = [None, u"qwe'qwe", 6.6]
    assert skytools.quote_statement(sql, args) == "set a=null, b='qwe''qwe', c='6.6'"

    sql2 = "set a=%(a)s, b=%(b)s, c=%(c)s"
    args2 = dict(a=None, b="qwe'qwe", c=6.6)
    assert skytools.quote_statement(sql2, args2) == "set a=null, b='qwe''qwe', c='6.6'"


def test_quote_json() -> None:
    json_string_vals = [
        (None, "null"),
        ('', '""'),
        (u'xx', '"xx"'),
        ('qwe"qwe\t', '"qwe\\"qwe\\t"'),
        ('\x01', '"\\u0001"'),
    ]
    try_quote(skytools.quote_json, json_string_vals)


def test_unquote_ident() -> None:
    idents = [
        ('qwe', 'qwe'),
        ('"qwe"', 'qwe'),
        ('"q""w\\\\e"', 'q"w\\\\e'),
        ('Foo', 'foo'),
        ('"Wei "" rd"', 'Wei " rd'),
    ]
    for val, exp in idents:
        assert skytools.unquote_ident(val) == exp



def test_unquote_ident_fail() -> None:
    with pytest.raises(Exception):
        skytools.unquote_ident('asd"asd')


def test_unescape_copy() -> None:
    assert unescape_copy(r'baz\tfo\'o') == "baz\tfo'o"
    assert unescape_copy(r'\N') is None


def test_unquote_fqident() -> None:
    assert unquote_fqident('Foo') == 'foo'
    assert unquote_fqident('"Foo"."Bar "" z"') == 'Foo.Bar " z'


def test_json_encode() -> None:
    assert json_encode({'a': 1}) == '{"a": 1}'
    assert json_encode('a') == '"a"'
    assert json_encode(['a']) == '["a"]'
    assert json_encode(a=1) == '{"a": 1}'


def test_json_decode() -> None:
    assert json_decode('[1]') == [1]


def test_make_pgarray() -> None:
    assert make_pgarray([]) == '{}'
    assert make_pgarray(['foo_3', 1, '', None]) == '{foo_3,1,"",NULL}'

    res = make_pgarray([None, ',', '\\', "'", '"', "{", "}", '_'])
    exp = '{NULL,",","\\\\","\'","\\"","{","}",_}'
    assert res == exp




"""Various parsers for Postgres-specific data formats.
"""

import re
from typing import Iterator, List, Optional, Sequence, Tuple, Dict, Union

import skytools

__all__ = (
    "parse_pgarray", "parse_logtriga_sql", "parse_tabbed_table",
    "parse_statements", 'sql_tokenizer', 'parse_sqltriga_sql',
    "parse_acl", "dedent", "hsize_to_bytes",
    "parse_connect_string", "merge_connect_string",
)

_rc_listelem = re.compile(r'( [^,"}]+ | ["] ( [^"\\]+ | [\\]. )* ["] )', re.X)


def parse_pgarray(array: Optional[str]) -> Optional[List[Optional[str]]]:
    r"""Parse Postgres array and return list of items inside it.
    """
    if array is None:
        return None
    if not array or array[0] not in ("{", "[") or array[-1] != '}':
        raise ValueError("bad array format: must be surrounded with {}")
    res = []
    pos = 1
    # skip optional dimensions descriptor "[a,b]={...}"
    if array[0] == "[":
        pos = array.find('{') + 1
        if pos < 1:
            raise ValueError("bad array format 2: must be surrounded with {}")
    while True:
        m = _rc_listelem.search(array, pos)
        if not m:
            break
        pos2 = m.end()
        item = array[pos:pos2]
        if len(item) == 4 and item.upper() == "NULL":
            val = None
        else:
            if item and item[0] == '"':
                item = item[1:-1]
            val = skytools.unescape(item)
        res.append(val)

        pos = pos2 + 1
        if array[pos2] == "}":
            break
        elif array[pos2] != ",":
            raise ValueError("bad array format: expected ,} got " + repr(array[pos2]))
    if pos < len(array) - 1:
        raise ValueError("bad array format: failed to parse completely (pos=%d len=%d)" % (pos, len(array)))
    return res


#
# parse logtriga partial sql
#

class _logtriga_parser:
    """Parses logtriga/sqltriga partial SQL to values."""
    pklist: List[str]
    def tokenizer(self, sql: str) -> Iterator[str]:
        """Token generator."""
        for tup in sql_tokenizer(sql, ignore_whitespace=True):
            yield tup[1]

    def parse_insert(self, tk: Iterator[str], fields: List[str], values: List[str],
                     key_fields: List[str], key_values: List[str]) -> None:
        """Handler for inserts."""
        # (col1, col2) values ('data', null)
        if next(tk) != "(":
            raise ValueError("syntax error")
        while True:
            fields.append(next(tk))
            t = next(tk)
            if t == ")":
                break
            elif t != ",":
                raise ValueError("syntax error")
        if next(tk).lower() != "values":
            raise ValueError("syntax error, expected VALUES")
        if next(tk) != "(":
            raise ValueError("syntax error, expected (")
        while True:
            values.append(next(tk))
            t = next(tk)
            if t == ")":
                break
            if t == ",":
                continue
            raise ValueError("expected , or ) got " + t)
        t = next(tk)
        raise ValueError("expected EOF, got " + repr(t))

    def parse_update(self, tk: Iterator[str], fields: List[str], values: List[str],
                     key_fields: List[str], key_values: List[str]) -> None:
        """Handler for updates."""
        # col1 = 'data1', col2 = null where pk1 = 'pk1' and pk2 = 'pk2'
        while True:
            fields.append(next(tk))
            if next(tk) != "=":
                raise ValueError("syntax error")
            values.append(next(tk))
            t = next(tk)
            if t == ",":
                continue
            elif t.lower() == "where":
                break
            else:
                raise ValueError("syntax error, expected WHERE or , got " + repr(t))
        while True:
            fld = next(tk)
            key_fields.append(fld)
            self.pklist.append(fld)
            if next(tk) != "=":
                raise ValueError("syntax error")
            key_values.append(next(tk))
            t = next(tk)
            if t.lower() != "and":
                raise ValueError("syntax error, expected AND got " + repr(t))

    def parse_delete(self, tk: Iterator[str], fields: List[str], values: List[str],
                     key_fields: List[str], key_values: List[str]) -> None:
        """Handler for deletes."""
        # pk1 = 'pk1' and pk2 = 'pk2'
        while True:
            fld = next(tk)
            key_fields.append(fld)
            self.pklist.append(fld)
            if next(tk) != "=":
                raise ValueError("syntax error")
            key_values.append(next(tk))
            t = next(tk)
            if t.lower() != "and":
                raise ValueError("syntax error, expected AND, got " + repr(t))

    def _create_dbdict(self, fields: List[str], values: List[str]) -> skytools.dbdict:
        fields2 = [skytools.unquote_ident(f) for f in fields]
        values2 = [skytools.unquote_literal(v) for v in values]
        return skytools.dbdict(zip(fields2, values2))

    def parse_sql(self, op: str, sql: str, pklist: Optional[Sequence[str]] = None, splitkeys: bool = False
    ) -> Union[skytools.dbdict, Tuple[skytools.dbdict, skytools.dbdict]]:
        """Main entry point."""
        if pklist is None:
            self.pklist = []
        else:
            self.pklist = list(pklist)
        tk = self.tokenizer(sql)
        fields: List[str] = []
        values: List[str] = []
        key_fields: List[str] = []
        key_values: List[str] = []
        try:
            if op == "I":
                self.parse_insert(tk, fields, values, key_fields, key_values)
            elif op == "U":
                self.parse_update(tk, fields, values, key_fields, key_values)
            elif op == "D":
                self.parse_delete(tk, fields, values, key_fields, key_values)
            raise ValueError("syntax error")
        except StopIteration:
            pass
        # last sanity check
        if (len(fields) + len(key_fields) == 0 or
            len(fields) != len(values) or
                len(key_fields) != len(key_values)):
            raise ValueError("syntax error, fields do not match values")
        if splitkeys:
            return (self._create_dbdict(key_fields, key_values),
                    self._create_dbdict(fields, values))
        return self._create_dbdict(fields + key_fields, values + key_values)


def parse_logtriga_sql(
    op: str, sql: str, splitkeys: bool = False
) -> Union[skytools.dbdict, Tuple[skytools.dbdict, skytools.dbdict]]:
    return parse_sqltriga_sql(op, sql, splitkeys=splitkeys)


def parse_sqltriga_sql(
    op: str, sql: str, pklist: Optional[Sequence[str]] = None, splitkeys: bool = False
) -> Union[skytools.dbdict, Tuple[skytools.dbdict, skytools.dbdict]]:
    """Parse partial SQL used by pgq.sqltriga() back to data values.

    Parser has following limitations:
     - Expects standard_quoted_strings = off
     - Does not support dollar quoting.
     - Does not support complex expressions anywhere. (hashtext(col1) = hashtext(val1))
     - WHERE expression must not contain IS (NOT) NULL
     - Does not support updating pk value, unless you use the splitkeys parameter.

    Returns dict of col->data pairs.
    """
    return _logtriga_parser().parse_sql(op, sql, pklist, splitkeys=splitkeys)


def parse_tabbed_table(txt: str) -> List[Dict[str, str]]:
    r"""Parse a tab-separated table into list of dicts.

    Expect first row to be column names.

    Very primitive.
    """

    txt = txt.replace("\r\n", "\n")
    fields = None
    data: List[Dict[str, str]] = []
    for ln in txt.split("\n"):
        if not ln:
            continue
        if not fields:
            fields = ln.split("\t")
            continue
        cols = ln.split("\t")
        if len(cols) != len(fields):
            continue
        row = dict(zip(fields, cols))
        data.append(row)
    return data


_extstr = r""" ['] (?: [^'\\]+ | \\. | [']['] )* ['] """
_stdstr = r""" ['] (?: [^']+ | [']['] )* ['] """
_name = r""" (?: [a-z_][a-z0-9_$]* | " (?: [^"]+ | "" )* " ) """

_ident = r""" (?P<ident> %s ) """ % _name
_fqident = r""" (?P<ident> %s (?: \. %s )* ) """ % (_name, _name)

_base_sql = r"""
      (?P<dolq>   (?P<dname> [$] (?: [_a-z][_a-z0-9]*)? [$] )
                  .*?
                  (?P=dname) )
    | (?P<num>    [0-9][0-9.e]* )
    | (?P<numarg> [$] [0-9]+ )
    | (?P<pyold>  [%][(] [a-z_][a-z0-9_]* [)] [s] )
    | (?P<pynew>  [{] [^{}]+ [}] )
    | (?P<ws>     (?: \s+ | [/][*] .*? [*][/] | [-][-][^\n]* )+ )
    | (?P<sym>    (?: [-+*~!@#^&|?/%<>=]+ | [,()\[\].:;] ) )
    | (?P<error>  . )"""

_base_sql_fq = r"%s | %s" % (_fqident, _base_sql)
_base_sql = r"%s | %s" % (_ident, _base_sql)

_std_sql = r"""(?: (?P<str> [E] %s | %s ) | %s )""" % (_extstr, _stdstr, _base_sql)
_std_sql_fq = r"""(?: (?P<str> [E] %s | %s ) | %s )""" % (_extstr, _stdstr, _base_sql_fq)
_ext_sql = r"""(?: (?P<str> [E]? %s ) | %s )""" % (_extstr, _base_sql)
_ext_sql_fq = r"""(?: (?P<str> [E]? %s ) | %s )""" % (_extstr, _base_sql_fq)
_std_sql_rc = _ext_sql_rc = None
_std_sql_fq_rc = _ext_sql_fq_rc = None


def sql_tokenizer(
        sql: str, standard_quoting: bool = False, ignore_whitespace: bool = False,
        fqident: bool = False, show_location: bool = False
        ) -> Iterator[Union[Tuple[str, str], Tuple[str, str, int]]]:
    r"""Parser SQL to tokens.

    Iterator, returns (toktype, tokstr) tuples.
    """
    global _std_sql_rc, _ext_sql_rc, _std_sql_fq_rc, _ext_sql_fq_rc
    if not _std_sql_rc:
        _std_sql_rc = re.compile(_std_sql, re.X | re.I | re.S)
        _ext_sql_rc = re.compile(_ext_sql, re.X | re.I | re.S)
        _std_sql_fq_rc = re.compile(_std_sql_fq, re.X | re.I | re.S)
        _ext_sql_fq_rc = re.compile(_ext_sql_fq, re.X | re.I | re.S)

    if standard_quoting:
        if fqident:
            rc = _std_sql_fq_rc
        else:
            rc = _std_sql_rc
    else:
        if fqident:
            rc = _ext_sql_fq_rc
        else:
            rc = _ext_sql_rc

    pos = 0
    while True:
        m = rc.match(sql, pos)
        if not m:
            break
        pos = m.end()
        typ = m.lastgroup or '?'
        if ignore_whitespace and typ == "ws":
            continue
        tk = m.group()
        if show_location:
            yield (typ, tk, pos)
        else:
            yield (typ, tk)


_copy_from_stdin_re = r"copy.*from\s+stdin"
_copy_from_stdin_rc = None


def parse_statements(sql: str, standard_quoting: bool = False) -> Iterator[str]:
    """Parse multi-statement string into separate statements.

    Returns list of statements.
    """

    global _copy_from_stdin_rc
    if not _copy_from_stdin_rc:
        _copy_from_stdin_rc = re.compile(_copy_from_stdin_re, re.X | re.I)
    tokens: List[str] = []
    pcount = 0  # '(' level
    for tmp in sql_tokenizer(sql, standard_quoting=standard_quoting):
        typ, t = tmp[0], tmp[1]
        # skip whitespace and comments before statement
        if len(tokens) == 0 and typ == "ws":
            continue
        # keep the rest
        tokens.append(t)
        if t == "(":
            pcount += 1
        elif t == ")":
            pcount -= 1
        elif t == ";" and pcount == 0:
            sql = "".join(tokens)
            if _copy_from_stdin_rc.match(sql):
                raise ValueError("copy from stdin not supported")
            yield "".join(tokens)
            tokens = []
    if len(tokens) > 0:
        yield "".join(tokens)
    if pcount != 0:
        raise ValueError("syntax error - unbalanced parenthesis")


_acl_name = r'(?: [0-9a-z_]+ | " (?: [^"]+ | "" )* " )'
_acl_re = r'''
    \s* (?: group \s+ | user \s+ )?
    (?P<tgt> %s )?
    (?P<perm> = [a-z*]*  )?
    (?P<owner> / %s )?
    \s* $
    ''' % (_acl_name, _acl_name)
_acl_rc = None


def parse_acl(acl: str) -> Optional[Tuple[Optional[str], str, Optional[str]]]:
    """Parse ACL entry.
    """
    global _acl_rc
    if not _acl_rc:
        _acl_rc = re.compile(_acl_re, re.I | re.X)

    m = _acl_rc.match(acl)
    if not m:
        return None

    target = m.group('tgt')
    perm = m.group('perm')
    owner = m.group('owner')

    if target:
        target = skytools.unquote_ident(target)
    if perm:
        perm = perm[1:]
    if owner:
        owner = skytools.unquote_ident(owner[1:])

    return (target, perm, owner)


def dedent(doc: str) -> str:
    r"""Relaxed dedent.

    - takes whitespace to be removed from first indented line.
    - allows empty or non-indented lines at the start
    - allows first line to be unindented
    - skips empty lines at the start
    - ignores indent of empty lines
    - if line does not match common indent, is stays unchanged
    """
    pfx: Optional[str] = None
    res: List[str] = []
    for ln in doc.splitlines():
        ln = ln.rstrip()
        if not pfx and len(res) < 2:
            if not ln:
                continue
            wslen = len(ln) - len(ln.lstrip())
            pfx = ln[: wslen]
        if pfx:
            if ln.startswith(pfx):
                ln = ln[len(pfx):]
        res.append(ln)
    res.append('')
    return '\n'.join(res)


def hsize_to_bytes(input_str: str) -> int:
    """ Convert sizes from human format to bytes (string to integer)
    """

    m = re.match(r"^([0-9]+) *([KMGTPEZY]?)B?$", input_str.strip(), re.IGNORECASE)
    if not m:
        raise ValueError("cannot parse: %s" % input_str)
    units = ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y']
    unit = m.group(2) or ''
    nbytes = int(m.group(1)) * 1024 ** units.index(unit.upper())
    assert isinstance(nbytes, int)
    return nbytes


#
# Connect string parsing
#

_cstr_rx = r""" \s* (\w+) \s* = \s* ( ' ( \\.| [^'\\] )* ' | \S+ ) \s* """
_cstr_unesc_rx = r"\\(.)"
_cstr_badval_rx = r"[\s'\\]"
_cstr_rc = None
_cstr_unesc_rc = None
_cstr_badval_rc = None


def parse_connect_string(cstr: str) -> List[Tuple[str, str]]:
    r"""Parse Postgres connect string.
    """
    global _cstr_rc, _cstr_unesc_rc
    if not _cstr_rc:
        _cstr_rc = re.compile(_cstr_rx, re.X)
        _cstr_unesc_rc = re.compile(_cstr_unesc_rx)
    pos = 0
    res = []
    while pos < len(cstr):
        m = _cstr_rc.match(cstr, pos)
        if not m:
            raise ValueError('Invalid connect string')
        pos = m.end()
        k = m.group(1)
        v = m.group(2)
        if v[0] == "'":
            v = _cstr_unesc_rc.sub(r"\1", v)
        res.append((k, v))
    return res


def merge_connect_string(cstr_arg_list: Sequence[Tuple[str, str]]) -> str:
    """Put fragments back together.
    """
    global _cstr_badval_rc
    if not _cstr_badval_rc:
        _cstr_badval_rc = re.compile(_cstr_badval_rx)

    buf = []
    for k, v in cstr_arg_list:
        if not v or _cstr_badval_rc.search(v):
            v = v.replace('\\', r'\\')
            v = v.replace("'", r"\'")
            v = "'" + v + "'"
        buf.append("%s=%s" % (k, v))
    return ' '.join(buf)





import pytest

from skytools.utf8 import safe_utf8_decode, sanitize_unicode


def test_safe_decode() -> None:
    assert safe_utf8_decode(b"foobar") == (True, "foobar")
    assert safe_utf8_decode(b'X\0Z') == (False, "X\uFFFDZ")
    assert safe_utf8_decode(b"OK") == (True, "OK")
    assert safe_utf8_decode(b'X\xF1Y') == (False, "X\uFFFDY")

    assert sanitize_unicode(u'\uD801\uDC01') == "\U00010401"

    with pytest.raises(TypeError):
        sanitize_unicode(b'qwe')    # type: ignore[arg-type]

## these give different results in py27 and py35
# >>> _norm(safe_utf8_decode(b'X\xed\xa0\x80Y\xed\xb0\x89Z'))
# (False, ['X', 65533, 65533, 65533, 'Y', 65533, 65533, 65533, 'Z'])
# >>> _norm(safe_utf8_decode(b'X\xed\xa0\x80\xed\xb0\x89Z'))
# (False, ['X', 65533, 65533, 65533, 65533, 65533, 65533, 'Z'])




"""Natural sort.

Compares numeric parts numerically.

Rules:

* String consists of numeric and non-numeric parts.
* Parts are compared pairwise, if both are numeric then numerically,
  otherwise textually.  (Only first fragment can have different type)
* If strings are different, they should compare differnet in natsort.

Extra rules for version numbers:

* In textual comparision, '~' is less than anything else, including end-of-string.
* In numeric comparision, numbers that start with '0' are compared as decimal
  fractions, thus number that starts with more zeroes is smaller.

"""

import re
from typing import List, Sequence

__all__ = (
    'natsort_key', 'natsort', 'natsorted',
    'natsort_key_icase', 'natsort_icase', 'natsorted_icase'
)

_rc = re.compile(r'\d+|\D+', re.A)


def natsort_key(s: str) -> str:
    """Returns string that sorts according to natsort rules.
    """
    # generates four types of fragments:
    # 1) strings < "0", stay as-is
    # 2) numbers starting with 0, fragment starts with "A".."Z"
    # 3) numbers starting with 1..9, fragment starts with "a".."z"
    # 4) strings > "9", fragment starts with "|"
    if "~" in s:
        s = s.replace("~", "\0")
    key: List[str] = []
    key_append = key.append
    for frag in _rc.findall(s):
        if frag < "0":
            key_append(frag)
            key_append("\1")
        elif frag < "1":
            nzeros = len(frag) - len(frag.lstrip('0'))
            mag = str(nzeros)
            mag = str(10**len(mag) - nzeros)
            key_append(chr(0x5B - len(mag)))    # Z, Y, X, ...
            key_append(mag)
            key_append(frag)
        elif frag < ":":
            mag = str(len(frag))
            key_append(chr(0x60 + len(mag)))    # a, b, c, ...
            key_append(mag)
            key_append(frag)
        else:
            key_append("|")
            key_append(frag)
            key_append("\1")
    if not (key and key[-1] == "\1"):
        key_append("\1")
    return "".join(key)


def natsort(lst: List[str]) -> None:
    """Natural in-place sort, case-sensitive."""
    lst.sort(key=natsort_key)


def natsorted(lst: Sequence[str]) -> List[str]:
    """Return copy of list, sorted in natural order, case-sensitive.
    """
    return sorted(lst, key=natsort_key)


# case-insensitive api

def natsort_key_icase(s: str) -> str:
    """Split string to numeric and non-numeric fragments."""
    return natsort_key(s.lower())


def natsort_icase(lst: List[str]) -> None:
    """Natural in-place sort, case-sensitive."""
    lst.sort(key=natsort_key_icase)


def natsorted_icase(lst: Sequence[str]) -> List[str]:
    """Return copy of list, sorted in natural order, case-sensitive.
    """
    return sorted(lst, key=natsort_key_icase)




"""Database tools.
"""

import io
import logging
import os
from typing import Any, Dict, List, Mapping, Optional, Sequence, Tuple, Union, cast, Callable

import skytools

from .basetypes import Connection, Cursor

__all__ = (
    "fq_name_parts", "fq_name", "get_table_oid", "get_table_pkeys",
    "get_table_columns", "exists_schema", "exists_table", "exists_type",
    "exists_sequence", "exists_temp_table", "exists_view",
    "exists_function", "exists_language", "Snapshot", "magic_insert",
    "CopyPipe", "full_copy", "DBObject", "DBSchema", "DBTable", "DBFunction",
    "DBLanguage", "db_install", "installer_find_file", "installer_apply_file",
    "dbdict", "mk_insert_sql", "mk_update_sql", "mk_delete_sql",
)


class dbdict(Dict[str, Any]):
    """Wrapper on actual dict that allows accessing dict keys as attributes.
    """
    # obj.foo access
    def __getattr__(self, k: str) -> Any:
        "Return attribute."
        try:
            return self[k]
        except KeyError:
            raise AttributeError(k) from None
    def __setattr__(self, k: str, v: Any) -> None:
        "Set attribute."
        self[k] = v
    def __delattr__(self, k: str) -> None:
        "Remove attribute."
        del self[k]
    def merge(self, other: Dict[str, Any]) -> None:
        for key in other:
            if key not in self:
                self[key] = other[key]


#
# Fully qualified table name
#

def fq_name_parts(tbl: str) -> Tuple[str, str]:
    """Return fully qualified name parts.
    """

    tmp = tbl.split('.', 1)
    if len(tmp) == 1:
        return ('public', tbl)
    return (tmp[0], tmp[1])


def fq_name(tbl: str) -> str:
    """Return fully qualified name.
    """
    return '.'.join(fq_name_parts(tbl))


#
# info about table
#

def get_table_oid(curs: Cursor, table_name: str) -> int:
    """Find Postgres OID for table."""
    schema, name = fq_name_parts(table_name)
    q = """select c.oid from pg_namespace n, pg_class c
           where c.relnamespace = n.oid
             and n.nspname = %s and c.relname = %s"""
    curs.execute(q, [schema, name])
    res = curs.fetchall()
    if len(res) == 0:
        raise Exception('Table not found: ' + table_name)
    return cast(int, res[0][0])


def get_table_pkeys(curs: Cursor, tbl: str) -> List[str]:
    """Return list of pkey column names."""
    oid = get_table_oid(curs, tbl)
    q = "SELECT k.attname FROM pg_index i, pg_attribute k"\
        " WHERE i.indrelid = %s AND k.attrelid = i.indexrelid"\
        "   AND i.indisprimary AND k.attnum > 0 AND NOT k.attisdropped"\
        " ORDER BY k.attnum"
    curs.execute(q, [oid])
    return [row[0] for row in curs.fetchall()]


def get_table_columns(curs: Cursor, tbl: str) -> List[str]:
    """Return list of column names for table."""
    oid = get_table_oid(curs, tbl)
    q = "SELECT k.attname FROM pg_attribute k"\
        " WHERE k.attrelid = %s"\
        "   AND k.attnum > 0 AND NOT k.attisdropped"\
        " ORDER BY k.attnum"
    curs.execute(q, [oid])
    return [row[0] for row in curs.fetchall()]


#
# exist checks
#

def exists_schema(curs: Cursor, schema: str) -> bool:
    """Does schema exists?"""
    q = "select count(1) from pg_namespace where nspname = %s"
    curs.execute(q, [schema])
    res = curs.fetchone()
    return bool(res[0])


def exists_table(curs: Cursor, table_name: str) -> bool:
    """Does table exists?"""
    schema, name = fq_name_parts(table_name)
    q = """select count(1) from pg_namespace n, pg_class c
           where c.relnamespace = n.oid and c.relkind = 'r'
             and n.nspname = %s and c.relname = %s"""
    curs.execute(q, [schema, name])
    res = curs.fetchone()
    return bool(res[0])


def exists_sequence(curs: Cursor, seq_name: str) -> bool:
    """Does sequence exists?"""
    schema, name = fq_name_parts(seq_name)
    q = """select count(1) from pg_namespace n, pg_class c
           where c.relnamespace = n.oid and c.relkind = 'S'
             and n.nspname = %s and c.relname = %s"""
    curs.execute(q, [schema, name])
    res = curs.fetchone()
    return bool(res[0])


def exists_view(curs: Cursor, view_name: str) -> bool:
    """Does view exists?"""
    schema, name = fq_name_parts(view_name)
    q = """select count(1) from pg_namespace n, pg_class c
           where c.relnamespace = n.oid and c.relkind = 'v'
             and n.nspname = %s and c.relname = %s"""
    curs.execute(q, [schema, name])
    res = curs.fetchone()
    return bool(res[0])


def exists_type(curs: Cursor, type_name: str) -> bool:
    """Does type exists?"""
    schema, name = fq_name_parts(type_name)
    q = """select count(1) from pg_namespace n, pg_type t
           where t.typnamespace = n.oid
             and n.nspname = %s and t.typname = %s"""
    curs.execute(q, [schema, name])
    res = curs.fetchone()
    return bool(res[0])


def exists_function(curs: Cursor, function_name: str, nargs: int) -> bool:
    """Does function exists?"""
    # this does not check arg types, so may match several functions
    schema, name = fq_name_parts(function_name)
    q = """select count(1) from pg_namespace n, pg_proc p
           where p.pronamespace = n.oid and p.pronargs = %s
             and n.nspname = %s and p.proname = %s"""
    curs.execute(q, [nargs, schema, name])
    res = curs.fetchone()

    # if unqualified function, check builtin functions too
    if not res[0] and function_name.find('.') < 0:
        name = "pg_catalog." + function_name
        return exists_function(curs, name, nargs)

    return bool(res[0])


def exists_language(curs: Cursor, lang_name: str) -> bool:
    """Does PL exists?"""
    q = """select count(1) from pg_language
           where lanname = %s"""
    curs.execute(q, [lang_name])
    res = curs.fetchone()
    return bool(res[0])


def exists_temp_table(curs: Cursor, tbl: str) -> bool:
    """Does temp table exists?"""
    # correct way, works only on 8.2
    q = "select 1 from pg_class where relname = %s and relnamespace = pg_my_temp_schema()"
    curs.execute(q, [tbl])
    tmp = curs.fetchall()
    return len(tmp) > 0


#
# Support for PostgreSQL snapshot
#

class Snapshot:
    """Represents a PostgreSQL snapshot.
    """

    def __init__(self, str_val: str):
        "Create snapshot from string."

        self.sn_str = str_val
        tmp = str_val.split(':')
        if len(tmp) != 3:
            raise ValueError('Unknown format for snapshot')
        self.xmin = int(tmp[0])
        self.xmax = int(tmp[1])
        self.txid_list = []
        if tmp[2] != "":
            for s in tmp[2].split(','):
                self.txid_list.append(int(s))

    def contains(self, txid: int) -> bool:
        "Is txid visible in snapshot."

        txid = int(txid)

        if txid < self.xmin:
            return True
        if txid >= self.xmax:
            return False
        if txid in self.txid_list:
            return False
        return True


#
# Copy helpers
#

def _gen_dict_copy(tbl: str, row: Mapping[str, Any], fields: Sequence[str], qfields: Sequence[str]) -> str:
    tmp: List[str] = []
    for f in fields:
        v = row.get(f)
        tmp.append(skytools.quote_copy(v))
    return "\t".join(tmp)


def _gen_dict_insert(tbl: str, row: Mapping[str, Any], fields: Sequence[str], qfields: Sequence[str]) -> str:
    tmp: List[str] = []
    for f in fields:
        v = row.get(f)
        tmp.append(skytools.quote_literal(v))
    fmt = "insert into %s (%s) values (%s);"
    return fmt % (tbl, ",".join(qfields), ",".join(tmp))


def _gen_list_copy(tbl: str, row: Sequence[Any], fields: Sequence[str], qfields: Sequence[str]) -> str:
    tmp: List[str] = []
    for i in range(len(fields)):
        try:
            v = row[i]
        except IndexError:
            v = None
        tmp.append(skytools.quote_copy(v))
    return "\t".join(tmp)


def _gen_list_insert(tbl: str, row: Sequence[Any], fields: Sequence[str], qfields: Sequence[str]) -> str:
    tmp: List[str] = []
    for i in range(len(fields)):
        try:
            v = row[i]
        except IndexError:
            v = None
        tmp.append(skytools.quote_literal(v))
    fmt = "insert into %s (%s) values (%s);"
    return fmt % (tbl, ",".join(qfields), ",".join(tmp))


DictRow = Mapping[str, Any]
ListRow = Sequence[Any]
DictRows = Sequence[DictRow]
ListRows = Sequence[ListRow]


def magic_insert(curs: Optional[Cursor], tablename: str,
                 data: Union[ListRows, DictRows],
                 fields: Optional[Sequence[str]] = None,
                 use_insert: bool = False, quoted_table: bool = False) -> Optional[str]:
    r"""Copy/insert a list of dict/list data to database.

    If curs is None, then the copy or insert statements are returned
    as string.  For list of dict the field list is optional, as its
    possible to guess them from dict keys.
    """
    if len(data) == 0:
        return None

    if fields is not None:
        fields = list(fields)  # get rid of iterator

    if quoted_table:
        qtablename = tablename
    else:
        qtablename = skytools.quote_fqident(tablename)

    # decide how to process
    if hasattr(data[0], 'keys'):
        if fields is None:
            fields = data[0].keys()     # type: ignore
        if use_insert:
            row_func = _gen_dict_insert
        else:
            row_func = _gen_dict_copy
    else:
        if fields is None:
            raise Exception("Non-dict data needs field list")
        if use_insert:
            row_func = _gen_list_insert     # type: ignore
        else:
            row_func = _gen_list_copy       # type: ignore

    qfields = [skytools.quote_ident(f) for f in fields]     # type: ignore

    # init processing
    buf = io.StringIO()
    if curs is None and use_insert == 0:
        fmt = "COPY %s (%s) FROM STDIN;\n"
        buf.write(fmt % (qtablename, ",".join(qfields)))

    # process data
    for row in data:
        buf.write(row_func(qtablename, row, fields, qfields))   # type: ignore
        buf.write("\n")

    # if user needs only string, return it
    if curs is None:
        if use_insert == 0:
            buf.write("\\.\n")
        return buf.getvalue()

    # do the actual copy/inserts
    if use_insert:
        curs.execute(buf.getvalue())
    else:
        buf.seek(0)
        sql = "COPY %s (%s) FROM STDIN" % (qtablename, ",".join(qfields))
        curs.copy_expert(sql, buf)
    return None


#
# Full COPY of table from one db to another
#

class CopyPipe(io.TextIOBase):
    """Splits one big COPY to chunks.
    """
    tablename: Optional[str]
    sql_from: Optional[str]
    dstcurs: Cursor
    buf: io.StringIO
    limit: int
    write_hook: Optional[Callable[["CopyPipe", str], str]]
    flush_hook: Optional[Callable[["CopyPipe"], None]]
    total_rows: int
    total_bytes: int

    def __init__(self, dstcurs: Cursor,
                 tablename: Optional[str] = None,
                 limit: int = 512 * 1024,
                 sql_from: Optional[str] = None):
        super().__init__()
        self.tablename = tablename
        self.sql_from = sql_from
        self.dstcurs = dstcurs
        self.buf = io.StringIO()
        self.limit = limit
        #hook for new data, hook func should return new data
        #def write_hook(obj, data):
        #   return data
        self.write_hook = None
        #hook for flush, hook func result is discarded
        # def flush_hook(obj):
        #   return None
        self.flush_hook = None
        self.total_rows = 0
        self.total_bytes = 0

    def write(self, data: str) -> int:
        """New row from psycopg
        """
        if self.write_hook:
            data = self.write_hook(self, data)

        self.total_bytes += len(data)  # it's chars now...
        self.total_rows += 1

        n = self.buf.write(data)

        if self.buf.tell() >= self.limit:
            self.flush()
        return n

    def flush(self) -> None:
        """Send data out.
        """
        if self.flush_hook:
            self.flush_hook(self)

        if self.buf.tell() <= 0:
            return

        self.buf.seek(0)
        if self.sql_from:
            sql = self.sql_from
        else:
            sql = "COPY %s FROM STDIN" % (self.tablename or "missing_table_name",)
        self.dstcurs.copy_expert(sql, self.buf)
        self.buf.seek(0)
        self.buf.truncate()


def full_copy(tablename: str, src_curs: Cursor, dst_curs: Cursor,
              column_list: Sequence[str] = (),
              condition: Optional[str] = None,
              dst_tablename: Optional[str] = None,
              dst_column_list: Optional[Sequence[str]] = None,
              write_hook: Optional[Callable[[CopyPipe, str], str]] = None,
              flush_hook: Optional[Callable[[CopyPipe], None]] = None) -> Tuple[int, int]:
    """COPY table from one db to another."""

    # default dst table and dst columns to source ones
    dst_tablename = dst_tablename or tablename
    dst_column_list = dst_column_list or column_list[:]
    if len(dst_column_list) != len(column_list):
        raise Exception('src and dst column lists must match in length')

    def build_qfields(cols: Sequence[str]) -> str:
        if cols:
            return ",".join([skytools.quote_ident(f) for f in cols])
        else:
            return "*"

    def build_statement(table: str, cols: Sequence[str]) -> str:
        qtable = skytools.quote_fqident(table)
        if cols:
            qfields = build_qfields(cols)
            return "%s (%s)" % (qtable, qfields)
        else:
            return qtable

    dst = build_statement(dst_tablename, dst_column_list)
    if condition:
        src = "(SELECT %s FROM ONLY %s WHERE %s)" % (build_qfields(column_list),
                                                skytools.quote_fqident(tablename),
                                                condition)
    else:
        src = build_statement(tablename, column_list)

    sql_to = "COPY %s TO stdout" % src
    sql_from = "COPY %s FROM stdin" % dst
    buf = CopyPipe(dst_curs, sql_from=sql_from)
    buf.write_hook = write_hook
    buf.flush_hook = flush_hook
    src_curs.copy_expert(sql_to, buf)
    buf.flush()

    return (buf.total_bytes, buf.total_rows)


#
# SQL installer
#

class DBObject:
    """Base class for installable DB objects."""
    name: str
    sql: Optional[str] = None
    sql_file: Optional[str] = None

    def __init__(self, name: str, sql: Optional[str] = None, sql_file: Optional[str] = None) -> None:
        """Generic dbobject init."""
        self.name = name
        self.sql = sql
        self.sql_file = sql_file

    def create(self, curs: Cursor, log: Optional[logging.Logger] = None) -> None:
        """Create a dbobject."""
        if log:
            log.info('Installing %s' % self.name)
        if self.sql:
            sql = self.sql
        elif self.sql_file:
            fn = self.find_file()
            if log:
                log.info("  Reading from %s" % fn)
            with open(fn, "r", encoding="utf8") as f:
                sql = f.read()
        else:
            raise Exception('object not defined')
        for stmt in skytools.parse_statements(sql):
            #if log: log.debug(repr(stmt))
            curs.execute(stmt)

    def find_file(self) -> str:
        """Find install script file."""
        if not self.sql_file:
            raise ValueError("sql_file not set")
        return installer_find_file(self.sql_file)

    def exists(self, curs: Cursor) -> int:
        raise NotImplementedError


class DBSchema(DBObject):
    """Handles db schema."""
    def exists(self, curs: Cursor) -> int:
        """Does schema exists."""
        return exists_schema(curs, self.name)


class DBTable(DBObject):
    """Handles db table."""
    def exists(self, curs: Cursor) -> int:
        """Does table exists."""
        return exists_table(curs, self.name)


class DBFunction(DBObject):
    """Handles db function."""

    def __init__(self, name: str, nargs: int, sql: Optional[str] = None, sql_file: Optional[str] = None) -> None:
        """Function object - number of args is significant."""
        super().__init__(name, sql, sql_file)
        self.nargs = nargs

    def exists(self, curs: Cursor) -> int:
        """Does function exists."""
        return exists_function(curs, self.name, self.nargs)


class DBLanguage(DBObject):
    """Handles db language."""

    def __init__(self, name: str) -> None:
        """PL object - creation happens with CREATE LANGUAGE."""
        super().__init__(name, sql="create language %s" % name)

    def exists(self, curs: Cursor) -> int:
        """Does PL exists."""
        return exists_language(curs, self.name)


def db_install(curs: Cursor, obj_list: Sequence[DBObject], log: Optional[logging.Logger] = None) -> None:
    """Installs list of objects into db."""
    for obj in obj_list:
        if not obj.exists(curs):
            obj.create(curs, log)
        else:
            if log:
                log.info('%s is installed' % obj.name)


def installer_find_file(filename: str) -> str:
    """Find SQL script from pre-defined paths."""
    full_fn = None
    if filename[0] == "/":
        if os.path.isfile(filename):
            full_fn = filename
    else:
        from skytools.installer_config import sql_locations
        dir_list = sql_locations
        for fdir in dir_list:
            fn = os.path.join(fdir, filename)
            if os.path.isfile(fn):
                full_fn = fn
                break

    if not full_fn:
        raise Exception('File not found: ' + filename)
    return full_fn


def installer_apply_file(db: Connection, filename: str, log: logging.Logger) -> None:
    """Find SQL file and apply it to db, statement-by-statement."""
    fn = installer_find_file(filename)
    with open(fn, "r", encoding="utf8") as f:
        sql = f.read()
    if log:
        log.info("applying %s" % fn)
    curs = db.cursor()
    for stmt in skytools.parse_statements(sql):
        #log.debug(repr(stmt))
        curs.execute(stmt)


#
# Generate INSERT/UPDATE/DELETE statement
#

def mk_insert_sql(row: DictRow, tbl: str,
                  pkey_list: Optional[Sequence[str]] = None,
                  field_map: Optional[Mapping[str, str]] = None) -> str:
    """Generate INSERT statement from dict data.
    """
    col_list = []
    val_list = []
    if field_map:
        for src, dst in field_map.items():
            col_list.append(skytools.quote_ident(dst))
            val_list.append(skytools.quote_literal(row[src]))
    else:
        for c, v in row.items():
            col_list.append(skytools.quote_ident(c))
            val_list.append(skytools.quote_literal(v))
    col_str = ", ".join(col_list)
    val_str = ", ".join(val_list)
    return "insert into %s (%s) values (%s);" % (
        skytools.quote_fqident(tbl), col_str, val_str)


def mk_update_sql(row: DictRow, tbl: str,
                  pkey_list: Sequence[str],
                  field_map: Optional[Mapping[str, str]] = None) -> str:
    """Generate UPDATE statement from dict data.
    """
    if len(pkey_list) < 1:
        raise Exception("update needs pkeys")
    set_list = []
    whe_list = []
    pkmap = {}
    for k in pkey_list:
        pkmap[k] = 1
        new_k = field_map and field_map[k] or k
        col = skytools.quote_ident(new_k)
        val = skytools.quote_literal(row[k])
        whe_list.append("%s = %s" % (col, val))

    if field_map:
        for src, dst in field_map.items():
            if src not in pkmap:
                col = skytools.quote_ident(dst)
                val = skytools.quote_literal(row[src])
                set_list.append("%s = %s" % (col, val))
    else:
        for col, val in row.items():
            if col not in pkmap:
                col = skytools.quote_ident(col)
                val = skytools.quote_literal(val)
                set_list.append("%s = %s" % (col, val))
    return "update only %s set %s where %s;" % (skytools.quote_fqident(tbl),
                                                ", ".join(set_list), " and ".join(whe_list))


def mk_delete_sql(row: DictRow, tbl: str,
                  pkey_list: Sequence[str],
                  field_map: Optional[Mapping[str, str]] = None) -> str:
    """Generate DELETE statement from dict data.
    """
    if len(pkey_list) < 1:
        raise Exception("delete needs pkeys")
    whe_list = []
    for k in pkey_list:
        new_k = field_map and field_map[k] or k
        col = skytools.quote_ident(new_k)
        val = skytools.quote_literal(row[k])
        whe_list.append("%s = %s" % (col, val))
    whe_str = " and ".join(whe_list)
    return "delete from only %s where %s;" % (skytools.quote_fqident(tbl), whe_str)




import os
import socket
import sys

import pytest

from skytools.sockutil import set_cloexec, set_nonblocking, set_tcp_keepalive


def test_set_tcp_keepalive() -> None:
    with socket.socket() as s:
        set_tcp_keepalive(s)


@pytest.mark.skipif(
    sys.platform == 'win32',
    reason="set_nonblocking on fd does not work on win32"
)
def test_set_nonblocking() -> None:
    with socket.socket() as s:
        assert set_nonblocking(s, None) == False
        assert set_nonblocking(s, True) == True
        assert set_nonblocking(s, None) == True


def test_set_cloexec_file() -> None:
    with open(os.devnull, 'rb') as f:
        assert set_cloexec(f, None) in (True, False)
        assert set_cloexec(f, True) == True
        assert set_cloexec(f, None) == True


def test_set_cloexec_socket() -> None:
    with socket.socket() as s:
        assert set_cloexec(s, None) in (True, False)
        assert set_cloexec(s, True) == True
        assert set_cloexec(s, None) == True





import io
import os.path
import sys

import pytest

from skytools.config import (
    Config, ConfigError, ExtendedCompatConfigParser,
    InterpolationError, NoOptionError, NoSectionError,
)

TOP = os.path.dirname(__file__)
CONFIG = os.path.join(TOP, 'config.ini')
CONFIG2 = os.path.join(TOP, 'config2.ini')


def test_config_str() -> None:
    cf = Config('base', CONFIG)
    assert cf.get('foo') == '1'
    assert cf.get('missing', 'q') == 'q'
    with pytest.raises(NoOptionError):
        cf.get('missing')


def test_config_int() -> None:
    cf = Config('base', CONFIG)
    assert cf.getint('foo') == 1
    assert cf.getint('missing', 2) == 2
    with pytest.raises(NoOptionError):
        cf.getint('missing')


def test_config_float() -> None:
    cf = Config('base', CONFIG)
    assert cf.getfloat('float-val') == 2.0
    assert cf.getfloat('missing', 3.0) == 3.0
    with pytest.raises(NoOptionError):
        cf.getfloat('missing')


def test_config_bool() -> None:
    cf = Config('base', CONFIG)
    assert cf.getboolean('bool-true1') == True
    assert cf.getboolean('bool-true2') == True
    assert cf.getboolean('missing', True) == True
    with pytest.raises(NoOptionError):
        cf.getboolean('missing')

    assert cf.getboolean('bool-false1') == False
    assert cf.getboolean('bool-false2') == False
    assert cf.getboolean('missing', False) == False
    with pytest.raises(NoOptionError):
        cf.getbool('missing')


def test_config_list() -> None:
    cf = Config('base', CONFIG)
    assert cf.getlist('list-val1') == []
    assert cf.getlist('list-val2') == ['a', '1', 'asd', 'ppp']
    assert cf.getlist('missing', ["a"]) == ["a"]
    with pytest.raises(NoOptionError):
        cf.getlist('missing')


def test_config_dict() -> None:
    cf = Config('base', CONFIG)
    assert cf.getdict('dict-val1') == {}
    assert cf.getdict('dict-val2') == {'a': '1', 'b': '2', 'z': 'z'}
    assert cf.getdict('missing', {'a': '1'}) == {'a': '1'}
    with pytest.raises(NoOptionError):
        cf.getdict('missing')


def test_config_file() -> None:
    cf = Config('base', CONFIG)
    assert cf.getfile('file-val1') == '-'
    if sys.platform != 'win32':
        assert cf.getfile('file-val2')[0] == '/'
    assert cf.getfile('missing', 'qwe') == 'qwe'
    with pytest.raises(NoOptionError):
        cf.getfile('missing')


def test_config_bytes() -> None:
    cf = Config('base', CONFIG)
    assert cf.getbytes('bytes-val1') == 4
    assert cf.getbytes('bytes-val2') == 2048
    assert cf.getbytes('missing', '3k') == 3072
    with pytest.raises(NoOptionError):
        cf.getbytes('missing')


def test_config_wildcard() -> None:
    cf = Config('base', CONFIG)

    assert cf.get_wildcard('wild-*-*', ['a', 'b']) == 'w.a.b'
    assert cf.get_wildcard('wild-*-*', ['a', 'x']) == 'w.a'
    assert cf.get_wildcard('wild-*-*', ['q', 'b']) == 'w2'
    assert cf.get_wildcard('missing-*-*', ['1', '2'], 'def') == 'def'
    with pytest.raises(NoOptionError):
        cf.get_wildcard('missing-*-*', ['1', '2'])


def test_config_default() -> None:
    cf = Config('base', CONFIG)
    assert cf.get('all') == 'yes'


def test_config_other() -> None:
    cf = Config('base', CONFIG)
    assert sorted(cf.sections()) == ['base', 'other', 'testscript']
    assert cf.has_section('base') == True
    assert cf.has_section('other') == True
    assert cf.has_section('missing') == False
    assert cf.has_section('DEFAULT') == False

    assert cf.has_option('missing') == False
    assert cf.has_option('all') == True
    assert cf.has_option('foo') == True

    cf2 = cf.clone('other')
    opts = list(sorted(cf2.options()))
    assert opts == [
        'all', 'config_dir', 'config_file', 'host_name',
        'job_name', 'service_name', 'test'
    ]
    assert len(cf2.items()) == len(cf2.options())


def test_loading() -> None:
    with pytest.raises(NoSectionError):
        Config('random', CONFIG)
    with pytest.raises(NoSectionError):
        Config('random', CONFIG)
    with pytest.raises(ConfigError):
        Config('random', 'random.ini')


def test_nofile() -> None:
    cf = Config('base', None, user_defs={'a': '1'})
    assert cf.sections() == ['base']
    assert cf.get('a') == '1'

    cf = Config('base', None, user_defs={'a': '1'}, ignore_defs=True)
    assert cf.get('a', '2') == '2'


def test_override() -> None:
    cf = Config('base', CONFIG, override={'foo': 'overrided'})
    assert cf.get('foo') == 'overrided'


def test_vars() -> None:
    cf = Config('base', CONFIG)
    assert cf.get('vars1') == 'V2=V3=Q3'

    with pytest.raises(InterpolationError):
        cf.get('bad1')


def test_extended_compat() -> None:
    config = u'[foo]\nkey = ${sub} $${nosub}\nsub = 2\n[bar]\nkey = ${foo:key}\n'
    cf = ExtendedCompatConfigParser()
    cf.read_file(io.StringIO(config), 'conf.ini')
    assert cf.get('bar', 'key') == '2 ${nosub}'

    config = u'[foo]\nloop1= ${loop1}\nloop2 = ${loop3}\nloop3 = ${loop2}\n'
    cf = ExtendedCompatConfigParser()
    cf.read_file(io.StringIO(config), 'conf.ini')
    with pytest.raises(InterpolationError):
        cf.get('foo', 'loop1')
    with pytest.raises(InterpolationError):
        cf.get('foo', 'loop2')

    config = u'[foo]\nkey = %(sub)s ${sub}\nsub = 2\n[bar]\nkey = %(foo:key)s\nkey2 = ${foo:key}\n'
    cf = ExtendedCompatConfigParser()
    cf.read_file(io.StringIO(config), 'conf.ini')
    assert cf.get('bar', 'key2') == '2 2'
    with pytest.raises(NoOptionError):
        cf.get('bar', 'key')

    config = u'[foo]\nkey = ${bad:xxx}\n[bad]\nsub = 1\n'
    cf = ExtendedCompatConfigParser()
    cf.read_file(io.StringIO(config), 'conf.ini')
    with pytest.raises(NoOptionError):
        cf.get('foo', 'key')


def test_config_format() -> None:
    cf1 = Config("fmt1", CONFIG2)
    cf2 = Config("fmt2", CONFIG2)

    with pytest.raises(ConfigError):
        Config("fmt3", CONFIG2)

    assert cf1.get("bar") == "1"
    assert cf2.get("bar1") == "%(foo)s"
    assert cf2.get("bar2") == "1"





import os
import signal
import sys
import time
import pathlib

import pytest

import skytools
from skytools.scripting import run_single_process

WIN32 = sys.platform == "win32"
CONF = os.path.join(os.path.dirname(__file__), "config.ini")

TEST_DB = os.environ.get("TEST_DB")


def checklog(log: str, word: str) -> bool:
    with open(log, 'r') as f:
        return word in f.read()


class Runner:
    def __init__(self, logfile: str, word: str, sleep: int = 0) -> None:
        self.logfile = logfile
        self.word = word
        self.sleep = sleep

    def run(self) -> None:
        with open(self.logfile, "a") as f:
            f.write(self.word + "\n")
        time.sleep(self.sleep)


@pytest.mark.skipif(WIN32, reason="cannot daemonize on win32")
def test_bg_process(tmp_path: pathlib.Path) -> None:
    pidfile = str(tmp_path / "proc.pid")
    logfile = str(tmp_path / "proc.log")

    run_single_process(Runner(logfile, "STEP1"), False, pidfile)
    while skytools.signal_pidfile(pidfile, 0):
        time.sleep(1)
    assert checklog(logfile, "STEP1")

    # daemonize from other process
    pid = os.fork()
    if pid == 0:
        run_single_process(Runner(logfile, "STEP2", 10), True, pidfile)

    time.sleep(2)
    with pytest.raises(SystemExit):
        run_single_process(Runner(logfile, "STEP3"), False, pidfile)
    skytools.signal_pidfile(pidfile, signal.SIGTERM)
    while skytools.signal_pidfile(pidfile, 0):
        time.sleep(1)

    assert checklog(logfile, "STEP2")
    assert not checklog(logfile, "STEP3")


class OptScript(skytools.BaseScript):
    ARGPARSE = False
    looping = 0

    def send_signal(self, code: int) -> None:
        print("signal: %s" % code)
        sys.exit(0)

    def work(self) -> None:
        print("opt=%s" % self.cf.get("opt"))


class ArgScript(OptScript):
    ARGPARSE = True


def test_optparse_script(capsys: pytest.CaptureFixture[str]) -> None:
    with pytest.raises(SystemExit):
        OptScript("testscript", ["-h"])
    res = capsys.readouterr()
    assert "display" in res.out


def test_argparse_script(capsys: pytest.CaptureFixture[str]) -> None:
    with pytest.raises(SystemExit):
        ArgScript("testscript", ["-h"])
    res = capsys.readouterr()
    assert "display" in res.out


@pytest.mark.skipif(WIN32, reason="use signals on win32")
def test_optparse_signals(capsys: pytest.CaptureFixture[str]) -> None:
    with pytest.raises(SystemExit):
        OptScript("testscript", ["-s", CONF])
    res = capsys.readouterr()
    assert "SIGINT" in res.out or f"signal: {signal.SIGINT}" in res.out

    with pytest.raises(SystemExit):
        OptScript("testscript", ["-r", CONF])
    res = capsys.readouterr()
    assert "SIGHUP" in res.out or f"signal: {signal.SIGHUP}" in res.out

    with pytest.raises(SystemExit):
        OptScript("testscript", ["-k", CONF])
    res = capsys.readouterr()
    assert "SIGTERM" in res.out or f"signal: {signal.SIGTERM}" in res.out


@pytest.mark.skipif(WIN32, reason="need to use signals")
def test_argparse_signals(capsys: pytest.CaptureFixture[str]) -> None:
    with pytest.raises(SystemExit):
        ArgScript("testscript", ["-s", CONF])
    res = capsys.readouterr()
    assert "SIGINT" in res.out or f"signal: {signal.SIGINT}" in res.out

    with pytest.raises(SystemExit):
        ArgScript("testscript", ["-r", CONF])
    res = capsys.readouterr()
    assert "SIGHUP" in res.out or f"signal: {signal.SIGHUP}" in res.out

    with pytest.raises(SystemExit):
        ArgScript("testscript", ["-k", CONF])
    res = capsys.readouterr()
    assert "SIGTERM" in res.out or f"signal: {signal.SIGTERM}" in res.out


def test_optparse_confopt(capsys: pytest.CaptureFixture[str]) -> None:
    s = ArgScript("testscript", [CONF])
    s.start()
    res = capsys.readouterr()
    assert "opt=test" in res.out


def test_argparse_confopt(capsys: pytest.CaptureFixture[str]) -> None:
    s = ArgScript("testscript", [CONF])
    s.start()
    res = capsys.readouterr()
    assert "opt=test" in res.out


class DBScript(skytools.DBScript):
    ARGPARSE = True
    looping = 0

    def work(self) -> None:
        db = self.get_database("db", connstr=TEST_DB)
        curs = db.cursor()
        curs.execute("select 1")
        curs.fetchall()
        self.close_database("db")
        print("OK")


@pytest.mark.skipif(not TEST_DB, reason="need database config")
def test_get_database(capsys: pytest.CaptureFixture[str]) -> None:
    s = DBScript("testscript", [CONF])
    s.start()
    res = capsys.readouterr()
    assert "OK" in res.out





"""SQL script locations."""

__all__ = ['sql_locations']

sql_locations = [
    "/usr/share/skytools3",
]

# PEP 440 version: [N!]N(.N)*[{a|b|rc}N][.postN][.devN]

package_version = "3.9.2"

skylog = 0




"""Various helpers for string quoting/unquoting.
"""

import json
import re
from typing import Any, Dict, Mapping, Match, Optional, Sequence, Union

try:
    from skytools._cquoting import (
        db_urldecode, db_urlencode, quote_bytea_raw,
        quote_copy, quote_literal, unescape, unquote_literal,
    )
except ImportError:
    from skytools._pyquoting import (
        db_urldecode, db_urlencode, quote_bytea_raw,
        quote_copy, quote_literal, unescape, unquote_literal,
    )

__all__ = (
    # _pyqoting / _cquoting
    "db_urldecode", "db_urlencode", "quote_bytea_raw",
    "quote_copy", "quote_literal", "unescape", "unquote_literal",
    # local
    "quote_bytea_literal", "quote_bytea_copy", "quote_statement",
    "quote_ident", "quote_fqident", "quote_json", "unescape_copy",
    "unquote_ident", "unquote_fqident",
    "json_encode", "json_decode",
    "make_pgarray",
)


#
# SQL quoting
#

def quote_bytea_literal(s: Optional[bytes]) -> str:
    """Quote bytea for regular SQL."""

    return quote_literal(quote_bytea_raw(s))


def quote_bytea_copy(s: Optional[bytes]) -> str:
    """Quote bytea for COPY."""

    return quote_copy(quote_bytea_raw(s))


def quote_statement(sql: str, dict_or_list: Union[Mapping[str, Any], Sequence[Any]]) -> str:
    """Quote whole statement.

    Data values are taken from dict or list or tuple.
    """
    if hasattr(dict_or_list, 'items'):
        qdict: Dict[str, str] = {}
        for k, v in dict_or_list.items():
            qdict[k] = quote_literal(v)
        return sql % qdict
    else:
        qvals = [quote_literal(v) for v in dict_or_list]
        return sql % tuple(qvals)


# reserved keywords (RESERVED_KEYWORD + TYPE_FUNC_NAME_KEYWORD + COL_NAME_KEYWORD)
# same list as postgres quote_ident()
_ident_kwmap = frozenset("""
    all analyse analyze and any array as asc asymmetric authorization between bigint binary
    bit boolean both case cast char character check coalesce collate collation column
    concurrently constraint create cross current_catalog current_date current_role current_schema
    current_time current_timestamp current_user dec decimal default deferrable desc distinct
    do else end except exists extract false fetch float for foreign freeze from full grant
    greatest group grouping having ilike in initially inner inout int integer intersect
    interval into is isnull join lateral leading least left like limit localtime localtimestamp
    national natural nchar none normalize not notnull null nullif numeric offset on only or order
    out outer overlaps overlay placing position precision primary real references returning
    right row select session_user setof similar smallint some substring symmetric table
    tablesample then time timestamp to trailing treat trim true union unique user using
    values varchar variadic verbose when where window with xmlattributes xmlconcat xmlelement
    xmlexists xmlforest xmlnamespaces xmlparse xmlpi xmlroot xmlserialize xmltable
""".split())

_ident_bad = re.compile(r"[^a-z0-9_]|^[0-9]")


def quote_ident(s: str) -> str:
    """Quote SQL identifier.

    If is checked against weird symbols and keywords.
    """

    if _ident_bad.search(s) or s in _ident_kwmap:
        s = '"%s"' % s.replace('"', '""')
    elif not s:
        return '""'
    return s


def quote_fqident(s: str) -> str:
    """Quote fully qualified SQL identifier.

    The '.' is taken as namespace separator and
    all parts are quoted separately
    """
    tmp = s.split('.', 1)
    if len(tmp) == 1:
        return 'public.' + quote_ident(s)
    return '.'.join([quote_ident(name) for name in tmp])


#
# quoting for JSON strings
#

_jsre = re.compile(r'[\x00-\x1F\\/"]')
_jsmap = {
    "\b": "\\b", "\f": "\\f", "\n": "\\n", "\r": "\\r",
    "\t": "\\t", "\\": "\\\\", '"': '\\"',
    "/": "\\/",   # to avoid html attacks
}


def _json_quote_char(m: Match[str]) -> str:
    """Quote single char."""
    c = m.group(0)
    try:
        return _jsmap[c]
    except KeyError:
        return r"\u%04x" % ord(c)


def quote_json(s: Optional[str]) -> str:
    """JSON style quoting."""
    if s is None:
        return "null"
    return '"%s"' % _jsre.sub(_json_quote_char, s)


def unescape_copy(val: str) -> Optional[str]:
    r"""Removes C-style escapes, also converts "\N" to None.
    """
    if val == r"\N":
        return None
    return unescape(val)


def unquote_ident(val: str) -> str:
    """Unquotes possibly quoted SQL identifier.
    """
    if len(val) > 1 and val[0] == '"' and val[-1] == '"':
        return val[1:-1].replace('""', '"')
    if val.find('"') > 0:
        raise Exception('unsupported syntax')
    return val.lower()


def unquote_fqident(val: str) -> str:
    """Unquotes fully-qualified possibly quoted SQL identifier.
    """
    tmp = val.split('.', 1)
    return '.'.join([unquote_ident(i) for i in tmp])


def json_encode(val: Any = None, **kwargs: Any) -> str:
    """Creates JSON string from Python object.
    """
    return json.dumps(val or kwargs)


def json_decode(s: str) -> Any:
    """Parses JSON string into Python object.
    """
    return json.loads(s)


#
# Create Postgres array
#

# any chars not in "good" set?  main bad ones: [ ,{}\"]
_pgarray_bad_rx = r"[^0-9a-z_.%&=()<>*/+-]"
_pgarray_bad_rc = re.compile(_pgarray_bad_rx)


def _quote_pgarray_elem(value: Any) -> str:
    if value is None:
        return 'NULL'
    s = str(value)
    if _pgarray_bad_rc.search(s):
        s = s.replace('\\', '\\\\')
        return '"' + s.replace('"', r'\"') + '"'
    elif not s:
        return '""'
    return s


def make_pgarray(lst: Sequence[Any]) -> str:
    r"""Formats Python list as Postgres array.
    Reverse of parse_pgarray().
    """
    items = [_quote_pgarray_elem(v) for v in lst]
    return '{' + ','.join(items) + '}'

